# Copyright 2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging
import operator
import shelve
import typing

from portage.package.ebuild.fetch import DistfileName

logger = logging.getLogger(__name__)


class ContentDB:
    """
    The content db serves to translate content digests to distfiles
    names, and distfiles names to content digests. All keys have one or
    more prefixes separated by colons. For a digest key, the first
    prefix is "digest" and the second prefix is the hash algorithm name.
    For a filename key, the prefix is "filename".

    The value associated with a digest key is a set of file names. The
    value associated with a distfile key is a set of content revisions.
    Each content revision is expressed as a dictionary of digests which
    is suitable for construction of a DistfileName instance.
    """

    def __init__(self, shelve_instance: shelve.Shelf):
        self._shelve = shelve_instance

    def add(self, filename: DistfileName):
        """
        Add file name and digests, creating a new content revision, or
        incrementing the reference count to an identical content revision
        if one exists. If the file name had previous content revisions,
        then they continue to exist independently of the new one.

        @param filename: file name with digests attribute
        """
        distfile_str = str(filename)
        distfile_key = "filename:{}".format(distfile_str)
        for k, v in filename.digests.items():
            if k != "size":
                digest_key = "digest:{}:{}".format(k.upper(), v.lower())
                try:
                    digest_files = self._shelve[digest_key]
                except KeyError:
                    digest_files = set()
                digest_files.add(distfile_str)
                self._shelve[digest_key] = digest_files
        try:
            content_revisions = self._shelve[distfile_key]
        except KeyError:
            content_revisions = set()

        revision_key = tuple(
            sorted(
                (
                    (algo.upper(), filename.digests[algo.upper()].lower())
                    for algo in filename.digests
                    if algo != "size"
                ),
                key=operator.itemgetter(0),
            )
        )
        content_revisions.add(revision_key)
        self._shelve[distfile_key] = content_revisions

    def remove(self, filename: DistfileName):
        """
        Remove a file name and digests from the database. If identical
        content is still referenced by one or more other file names,
        then those references are preserved (like removing one of many
        hardlinks). Also, this file name may reference other content
        revisions with different digests, and those content revisions
        will remain as well.

        @param filename: file name with digests attribute
        """
        distfile_key = "filename:{}".format(filename)
        try:
            content_revisions = self._shelve[distfile_key]
        except KeyError:
            pass
        else:
            remaining = set()
            for revision_key in content_revisions:
                if not any(
                    digest_item in revision_key
                    for digest_item in filename.digests.items()
                ):
                    remaining.add(revision_key)
                    continue
                for k, v in revision_key:
                    digest_key = "digest:{}:{}".format(k, v)
                    try:
                        digest_files = self._shelve[digest_key]
                    except KeyError:
                        digest_files = set()

                    try:
                        digest_files.remove(filename)
                    except KeyError:
                        pass

                    if digest_files:
                        self._shelve[digest_key] = digest_files
                    else:
                        try:
                            del self._shelve[digest_key]
                        except KeyError:
                            pass

            if remaining:
                logger.debug(("drop '%s' revision(s) from content db") % filename)
                self._shelve[distfile_key] = remaining
            else:
                logger.debug(("drop '%s' from content db") % filename)
                try:
                    del self._shelve[distfile_key]
                except KeyError:
                    pass

    def get_filenames_translate(
        self, filename: typing.Union[str, DistfileName]
    ) -> typing.Generator[DistfileName, None, None]:
        """
        Translate distfiles content digests to zero or more distfile names.
        If filename is already a distfile name, then it will pass
        through unchanged.

        A given content digest will translate to multiple distfile names if
        multiple associations have been created via the add method. The
        relationship between a content digest and a distfile name is similar
        to the relationship between an inode and a hardlink.

        @param filename: A filename listed by layout get_filenames
        """
        if not isinstance(filename, DistfileName):
            filename = DistfileName(filename)

        # Match content digests with zero or more content revisions.
        matched_revisions = {}

        for k, v in filename.digests.items():
            digest_item = (k.upper(), v.lower())
            digest_key = "digest:{}:{}".format(*digest_item)
            try:
                digest_files = self._shelve[digest_key]
            except KeyError:
                continue

            for distfile_str in digest_files:
                matched_revisions.setdefault(distfile_str, set())
                try:
                    content_revisions = self._shelve["filename:{}".format(distfile_str)]
                except KeyError:
                    pass
                else:
                    for revision_key in content_revisions:
                        if (
                            digest_item in revision_key
                            and revision_key not in matched_revisions[distfile_str]
                        ):
                            matched_revisions[distfile_str].add(revision_key)
                            yield DistfileName(distfile_str, digests=dict(revision_key))

        if not any(matched_revisions.values()):
            # Since filename matched zero content revisions, allow
            # it to pass through unchanged (on the path toward deletion).
            yield filename

    def __len__(self):
        return len(self._shelve)

    def __contains__(self, k):
        return k in self._shelve

    def __iter__(self):
        return self._shelve.__iter__()

    def items(self):
        return self._shelve.items()

    def __setitem__(self, k, v):
        self._shelve[k] = v

    def __getitem__(self, k):
        return self._shelve[k]

    def __delitem__(self, k):
        del self._shelve[k]

    def get(self, k, *args):
        return self._shelve.get(k, *args)

    def close(self):
        self._shelve.close()

    def clear(self):
        self._shelve.clear()



# Copyright 1999-2009 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.DepPriority import DepPriority
from _emerge.Package import Package


def _find_deep_system_runtime_deps(graph):
    deep_system_deps = set()
    node_stack = []
    for node in graph:
        if not isinstance(node, Package) or node.operation == "uninstall":
            continue
        if node.root_config.sets["system"].findAtomForPackage(node):
            node_stack.append(node)

    def ignore_priority(priority):
        """
        Ignore non-runtime priorities.
        """
        if isinstance(priority, DepPriority) and (
            priority.runtime or priority.runtime_post
        ):
            return False
        return True

    while node_stack:
        node = node_stack.pop()
        if node in deep_system_deps:
            continue
        deep_system_deps.add(node)
        for child in graph.child_nodes(node, ignore_priority=ignore_priority):
            if not isinstance(child, Package) or child.operation == "uninstall":
                continue
            node_stack.append(child)

    return deep_system_deps



# Copyright 2012-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SlotAbiDowngradeTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        super(SlotAbiDowngradeTestCase, self).__init__(*args, **kwargs)

    def testSubSlot(self):
        ebuilds = {
            "dev-libs/icu-4.8": {"EAPI": "5", "SLOT": "0/48"},
            "dev-libs/libxml2-2.7.8": {
                "EAPI": "5",
                "DEPEND": "dev-libs/icu:=",
                "RDEPEND": "dev-libs/icu:=",
            },
        }
        binpkgs = {
            "dev-libs/icu-49": {"EAPI": "5", "SLOT": "0/49"},
            "dev-libs/icu-4.8": {"EAPI": "5", "SLOT": "0/48"},
            "dev-libs/libxml2-2.7.8": {
                "EAPI": "5",
                "DEPEND": "dev-libs/icu:0/49=",
                "RDEPEND": "dev-libs/icu:0/49=",
            },
        }
        installed = {
            "dev-libs/icu-49": {"EAPI": "5", "SLOT": "0/49"},
            "dev-libs/libxml2-2.7.8": {
                "EAPI": "5",
                "DEPEND": "dev-libs/icu:0/49=",
                "RDEPEND": "dev-libs/icu:0/49=",
            },
        }

        world = ["dev-libs/libxml2"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True},
                success=True,
                mergelist=["dev-libs/icu-4.8", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True, "--ignore-built-slot-operator-deps": "y"},
                success=True,
                mergelist=["dev-libs/icu-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/icu-4.8", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={"--oneshot": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]dev-libs/icu-49"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-libs/icu-4.8", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["dev-libs/icu-4.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/icu-4.8", "dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkgonly": True},
                success=True,
                mergelist=[],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()

    def testWholeSlotSubSlotMix(self):
        ebuilds = {
            "dev-libs/glib-1.2.10": {"SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2=",
                "RDEPEND": "dev-libs/glib:2=",
            },
        }
        binpkgs = {
            "dev-libs/glib-1.2.10": {"SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/glib-2.32.3": {"EAPI": "5", "SLOT": "2/2.32"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2/2.32=",
                "RDEPEND": "dev-libs/glib:2/2.32=",
            },
        }
        installed = {
            "dev-libs/glib-1.2.10": {"EAPI": "5", "SLOT": "1"},
            "dev-libs/glib-2.32.3": {"EAPI": "5", "SLOT": "2/2.32"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2/2.32=",
                "RDEPEND": "dev-libs/glib:2/2.32=",
            },
        }

        world = ["dev-libs/glib:1", "dev-libs/dbus-glib"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True},
                success=True,
                mergelist=["dev-libs/glib-2.30.2", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True, "--ignore-built-slot-operator-deps": "y"},
                success=True,
                mergelist=["dev-libs/glib-2.30.2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/glib-2.30.2", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/glib"],
                options={"--oneshot": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]dev-libs/glib-2.32.3"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-libs/glib-2.30.2", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=["dev-libs/glib-2.30.2"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/glib-2.30.2", "dev-libs/dbus-glib-0.98"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkgonly": True},
                success=True,
                mergelist=[],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



#!/usr/bin/env python
# Copyright 1998-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

try:
    from setuptools.core import setup, Command, Extension
    from setuptools.command.build import build
    from setuptools.command.build_ext import build_ext as _build_ext
    from setuptools.command.build_scripts import build_scripts
    from setuptools.command.clean import clean
    from setuptools.command.install import install
    from setuptools.command.install_data import install_data
    from setuptools.command.install_lib import install_lib
    from setuptools.command.install_scripts import install_scripts
    from setuptools.command.sdist import sdist
    from setuptools.dep_util import newer
    from setuptools.dir_util import mkpath, remove_tree, copy_tree
    from setuptools.util import change_root, subst_vars
except ImportError:
    from distutils.core import setup, Command, Extension
    from distutils.command.build import build
    from distutils.command.build_ext import build_ext as _build_ext
    from distutils.command.build_scripts import build_scripts
    from distutils.command.clean import clean
    from distutils.command.install import install
    from distutils.command.install_data import install_data
    from distutils.command.install_lib import install_lib
    from distutils.command.install_scripts import install_scripts
    from distutils.command.sdist import sdist
    from distutils.dep_util import newer
    from distutils.dir_util import mkpath, remove_tree, copy_tree
    from distutils.util import change_root, subst_vars

import codecs
import collections
import glob
import itertools
import os
import os.path
import platform
import re
import subprocess
import sys

autodetect_pip = os.path.basename(os.environ.get("_", "")) == "pip" or os.path.basename(
    os.path.dirname(__file__)
).startswith("pip-")
venv_prefix = "" if sys.prefix == sys.base_prefix else sys.prefix
create_entry_points = bool(autodetect_pip or venv_prefix)
with open(os.path.join(os.path.dirname(__file__), "README.md"), "rt") as f:
    long_description = f.read()

# TODO:
# - smarter rebuilds of docs w/ 'install_docbook' and 'install_apidoc'.

# Dictionary of scripts.  The structure is
#   key   = location in filesystem to install the scripts
#   value = list of scripts, path relative to top source directory
x_scripts = {
    "bin": [
        "bin/ebuild",
        "bin/egencache",
        "bin/emerge",
        "bin/emerge-webrsync",
        "bin/emirrordist",
        "bin/glsa-check",
        "bin/portageq",
        "bin/quickpkg",
        "bin/gpkg-sign",
    ],
    "sbin": [
        "bin/archive-conf",
        "bin/dispatch-conf",
        "bin/emaint",
        "bin/env-update",
        "bin/etc-update",
        "bin/fixpackages",
        "bin/regenworld",
    ],
}

# Dictionary custom modules written in C/C++ here.  The structure is
#   key   = module name
#   value = list of C/C++ source code, path relative to top source directory
x_c_helpers = {
    "portage.util.libc": [
        "src/portage_util_libc.c",
    ],
}

if platform.system() == "Linux":
    x_c_helpers.update(
        {
            "portage.util.file_copy.reflink_linux": [
                "src/portage_util_file_copy_reflink_linux.c",
            ],
        }
    )


class x_build(build):
    """Build command with extra build_man call."""

    def run(self):
        build.run(self)
        self.run_command("build_man")


class build_man(Command):
    """Perform substitutions in manpages."""

    user_options = []

    def initialize_options(self):
        self.build_base = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_base", "build_base"))

    def run(self):
        for d, files in self.distribution.data_files:
            if not d.startswith("$mandir/"):
                continue

            for source in files:
                target = os.path.join(self.build_base, source)
                mkpath(os.path.dirname(target))

                if not newer(source, target) and not newer(__file__, target):
                    continue

                print("copying and updating %s -> %s" % (source, target))

                with codecs.open(source, "r", "utf8") as f:
                    data = f.readlines()
                data[0] = data[0].replace("VERSION", self.distribution.get_version())
                with codecs.open(target, "w", "utf8") as f:
                    f.writelines(data)


class docbook(Command):
    """Build docs using docbook."""

    user_options = [
        (
            "doc-formats=",
            None,
            "Documentation formats to build (all xmlto formats for docbook are allowed, comma-separated",
        ),
    ]

    def initialize_options(self):
        self.doc_formats = "xhtml,xhtml-nochunks"

    def finalize_options(self):
        self.doc_formats = self.doc_formats.replace(",", " ").split()

    def run(self):
        if not os.path.isdir("doc/fragment"):
            mkpath("doc/fragment")

        with open("doc/fragment/date", "w"):
            pass
        with open("doc/fragment/version", "w") as f:
            f.write("<releaseinfo>%s</releaseinfo>" % self.distribution.get_version())

        for f in self.doc_formats:
            print("Building docs in %s format..." % f)
            subprocess.check_call(
                ["xmlto", "-o", "doc", "-m", "doc/custom.xsl", f, "doc/portage.docbook"]
            )


class apidoc(Command):
    """Build API docs using apidoc."""

    user_options = []

    def initialize_options(self):
        self.build_lib = None

    def finalize_options(self):
        self.set_undefined_options("build_py", ("build_lib", "build_lib"))

    def run(self):
        self.run_command("build_py")

        print("Building API documentation...")

        process_env = os.environ.copy()
        pythonpath = self.build_lib
        try:
            pythonpath += ":" + process_env["PYTHONPATH"]
        except KeyError:
            pass
        process_env["PYTHONPATH"] = pythonpath

        subprocess.check_call(["make", "-C", "doc/api", "html"], env=process_env)


class install_docbook(install_data):
    """install_data for docbook docs"""

    user_options = install_data.user_options + [
        ("htmldir=", None, "HTML documentation install directory"),
    ]

    def initialize_options(self):
        install_data.initialize_options(self)
        self.htmldir = None

    def finalize_options(self):
        self.set_undefined_options("install", ("htmldir", "htmldir"))
        install_data.finalize_options(self)

    def run(self):
        if not os.path.exists("doc/portage.html"):
            self.run_command("docbook")
        self.data_files = [
            (self.htmldir, glob.glob("doc/*.html")),
        ]
        install_data.run(self)


class install_apidoc(install_data):
    """install_data for apidoc docs"""

    user_options = install_data.user_options + [
        ("htmldir=", None, "HTML documentation install directory"),
    ]

    def initialize_options(self):
        install_data.initialize_options(self)
        self.htmldir = None

    def finalize_options(self):
        self.set_undefined_options("install", ("htmldir", "htmldir"))
        install_data.finalize_options(self)

    def run(self):
        if not os.path.exists("doc/api/build/html/index.html"):
            self.run_command("apidoc")
        self.data_files = [
            (
                os.path.join(self.htmldir, "api"),
                glob.glob("doc/api/build/html/*.html")
                + glob.glob("doc/api/build/html/*.js"),
            ),
            (
                os.path.join(self.htmldir, "api/_static"),
                glob.glob("doc/api/build/html/_static/*"),
            ),
        ]
        install_data.run(self)


class x_build_scripts_custom(build_scripts):
    def finalize_options(self):
        build_scripts.finalize_options(self)
        if "dir_name" in dir(self):
            self.build_dir = os.path.join(self.build_dir, self.dir_name)
            if self.dir_name in x_scripts:
                self.scripts = x_scripts[self.dir_name]
            else:
                self.scripts = set(self.scripts)
                if not (create_entry_points and self.dir_name == "portage"):
                    for other_files in x_scripts.values():
                        self.scripts.difference_update(other_files)

    def run(self):
        # group scripts by subdirectory
        split_scripts = collections.defaultdict(list)
        for f in self.scripts:
            dir_name = os.path.dirname(f[len("bin/") :])
            split_scripts[dir_name].append(f)

        base_dir = self.build_dir
        base_scripts = self.scripts
        for d, files in split_scripts.items():
            self.build_dir = os.path.join(base_dir, d)
            self.scripts = files
            self.copy_scripts()

        # restore previous values
        self.build_dir = base_dir
        self.scripts = base_scripts


class x_build_scripts_bin(x_build_scripts_custom):
    dir_name = "bin"


class x_build_scripts_sbin(x_build_scripts_custom):
    dir_name = "sbin"


class x_build_scripts_portagebin(x_build_scripts_custom):
    dir_name = "portage"


class x_build_scripts(build_scripts):
    def initialize_option(self):
        build_scripts.initialize_options(self)

    def finalize_options(self):
        build_scripts.finalize_options(self)

    def run(self):
        self.run_command("build_scripts_bin")
        self.run_command("build_scripts_portagebin")
        self.run_command("build_scripts_sbin")


class x_clean(clean):
    """clean extended for doc & post-test cleaning"""

    @staticmethod
    def clean_docs():
        def get_doc_outfiles():
            for dirpath, _dirnames, filenames in os.walk("doc"):
                for f in filenames:
                    if f.endswith(".docbook") or f == "custom.xsl":
                        pass
                    else:
                        yield os.path.join(dirpath, f)

                # do not recurse
                break

        for f in get_doc_outfiles():
            print("removing %s" % repr(f))
            os.remove(f)

        if os.path.isdir("doc/fragment"):
            remove_tree("doc/fragment")

        if os.path.isdir("doc/api/build"):
            remove_tree("doc/api/build")

    def clean_tests(self):
        # do not remove incorrect dirs accidentally
        top_dir = os.path.normpath(os.path.join(self.build_lib, ".."))
        cprefix = os.path.commonprefix((self.build_base, top_dir))
        if cprefix != self.build_base:
            return

        bin_dir = os.path.join(top_dir, "bin")
        if os.path.exists(bin_dir):
            remove_tree(bin_dir)

        conf_dir = os.path.join(top_dir, "cnf")
        if os.path.islink(conf_dir):
            print("removing %s symlink" % repr(conf_dir))
            os.unlink(conf_dir)

        pni_file = os.path.join(top_dir, ".portage_not_installed")
        if os.path.exists(pni_file):
            print("removing %s" % repr(pni_file))
            os.unlink(pni_file)

    def clean_man(self):
        man_dir = os.path.join(self.build_base, "man")
        if os.path.exists(man_dir):
            remove_tree(man_dir)

    def run(self):
        if self.all:
            self.clean_tests()
            self.clean_docs()
            self.clean_man()

        clean.run(self)


class x_install(install):
    """install command with extra Portage paths"""

    user_options = install.user_options + [
        # note: $prefix and $exec_prefix are reserved for Python install
        ("system-prefix=", None, "Prefix for architecture-independent data"),
        ("system-exec-prefix=", None, "Prefix for architecture-specific data"),
        ("bindir=", None, "Install directory for main executables"),
        ("datarootdir=", None, "Data install root directory"),
        ("docdir=", None, "Documentation install directory"),
        ("htmldir=", None, "HTML documentation install directory"),
        ("mandir=", None, "Manpage root install directory"),
        ("portage-base=", "b", "Portage install base"),
        (
            "portage-bindir=",
            None,
            "Install directory for Portage internal-use executables",
        ),
        ("portage-datadir=", None, "Install directory for data files"),
        ("sbindir=", None, "Install directory for superuser-intended executables"),
        ("sysconfdir=", None, "System configuration path"),
    ]

    # note: the order is important for proper substitution
    paths = [
        ("system_prefix", "/usr"),
        ("system_exec_prefix", "$system_prefix"),
        ("bindir", "$system_exec_prefix/bin"),
        ("sbindir", "$system_exec_prefix/sbin"),
        ("sysconfdir", "/etc"),
        ("datarootdir", "$system_prefix/share"),
        ("docdir", "$datarootdir/doc/$package-$version"),
        ("htmldir", "$docdir/html"),
        ("mandir", "$datarootdir/man"),
        ("portage_base", "$system_exec_prefix/lib/portage"),
        ("portage_bindir", "$portage_base/bin"),
        ("portage_datadir", "$datarootdir/portage"),
        # not customized at the moment
        ("logrotatedir", "$sysconfdir/logrotate.d"),
        ("portage_confdir", "$portage_datadir/config"),
        ("portage_setsdir", "$portage_confdir/sets"),
    ]

    def initialize_options(self):
        install.initialize_options(self)

        for key, default in self.paths:
            setattr(self, key, default)
        self.subst_paths = {}

    def finalize_options(self):
        install.finalize_options(self)

        # substitute variables
        new_paths = {
            "package": self.distribution.get_name(),
            "version": self.distribution.get_version(),
        }
        for key, _default in self.paths:
            new_paths[key] = subst_vars(getattr(self, key), new_paths)
            setattr(self, key, new_paths[key])
        self.subst_paths = new_paths


class x_install_data(install_data):
    """install_data with customized path support"""

    user_options = install_data.user_options

    def initialize_options(self):
        install_data.initialize_options(self)
        self.build_base = None
        self.paths = None

    def finalize_options(self):
        install_data.finalize_options(self)
        self.set_undefined_options("build", ("build_base", "build_base"))
        self.set_undefined_options("install", ("subst_paths", "paths"))

    def run(self):
        def re_sub_file(path, pattern, repl):
            print("Rewriting %s" % path)
            with codecs.open(path, "r", "utf-8") as f:
                data = f.read()
            data = re.sub(pattern, repl, data, flags=re.MULTILINE)
            with codecs.open(path, "w", "utf-8") as f:
                f.write(data)

        if create_entry_points:
            re_sub_file("cnf/repos.conf", r"= /", "= %(EPREFIX)s/")
            re_sub_file("cnf/make.globals", r'DIR="/', 'DIR="${EPREFIX}/')

        self.run_command("build_man")

        def process_data_files(df):
            for d, files in df:
                # substitute man sources
                if d.startswith("$mandir/"):
                    files = [os.path.join(self.build_base, v) for v in files]

                # substitute variables in path
                d = subst_vars(d, self.paths)
                yield (d, files)

        old_data_files = self.data_files
        self.data_files = process_data_files(self.data_files)

        install_data.run(self)
        self.data_files = old_data_files


class x_install_lib(install_lib):
    """install_lib command with Portage path substitution"""

    user_options = install_lib.user_options

    def initialize_options(self):
        install_lib.initialize_options(self)
        self.portage_base = None
        self.portage_bindir = None
        self.portage_confdir = None

    def finalize_options(self):
        install_lib.finalize_options(self)
        self.set_undefined_options(
            "install",
            ("portage_base", "portage_base"),
            ("portage_bindir", "portage_bindir"),
            ("portage_confdir", "portage_confdir"),
        )

    def install(self):
        ret = install_lib.install(self)

        def rewrite_file(path, val_dict):
            path = os.path.join(self.install_dir, path)
            print("Rewriting %s" % path)
            with codecs.open(path, "r", "utf-8") as f:
                data = f.read()

            for varname, val in val_dict.items():
                regexp = r"(?m)^(%s\s*=).*$" % varname
                repl = r"\1 %s" % repr(val)

                data = re.sub(regexp, repl, data)

            with codecs.open(path, "w", "utf-8") as f:
                f.write(data)

        rewrite_file(
            "portage/__init__.py",
            {
                "VERSION": self.distribution.get_version(),
            },
        )

        def re_sub_file(path, pattern_repl_items):
            path = os.path.join(self.install_dir, path)
            print("Rewriting %s" % path)
            with codecs.open(path, "r", "utf-8") as f:
                data = f.read()
            for pattern, repl in pattern_repl_items:
                data = re.sub(pattern, repl, data, flags=re.MULTILINE)
            with codecs.open(path, "w", "utf-8") as f:
                f.write(data)

        val_dict = {}
        if create_entry_points:
            re_sub_file(
                "portage/const.py",
                (
                    (
                        r"^(GLOBAL_CONFIG_PATH\s*=\s*[\"'])(.*)([\"'])",
                        lambda m: "{}{}{}".format(
                            m.group(1),
                            m.group(2).partition("/usr")[-1],
                            m.group(3),
                        ),
                    ),
                    (
                        r"^(PORTAGE_BASE_PATH\s*=\s*)(.*)",
                        lambda m: "{}{}".format(
                            m.group(1),
                            'os.path.join(os.path.realpath(__import__("sys").prefix), "lib/portage")',
                        ),
                    ),
                    (
                        r"^(EPREFIX\s*=\s*)(.*)",
                        lambda m: "{}{}".format(
                            m.group(1),
                            '__import__("sys").prefix',
                        ),
                    ),
                ),
            )
        else:
            val_dict.update(
                {
                    "PORTAGE_BASE_PATH": self.portage_base,
                    "PORTAGE_BIN_PATH": self.portage_bindir,
                }
            )
        rewrite_file("portage/const.py", val_dict)

        return ret


class x_install_scripts_custom(install_scripts):
    def initialize_options(self):
        install_scripts.initialize_options(self)
        self.root = None

    def finalize_options(self):
        self.set_undefined_options(
            "install", ("root", "root"), (self.var_name, "install_dir")
        )
        install_scripts.finalize_options(self)
        self.build_dir = os.path.join(self.build_dir, self.dir_name)

        # prepend root
        if self.root is not None:
            self.install_dir = change_root(self.root, self.install_dir)

    def run(self):
        if not create_entry_points:
            install_scripts.run(self)


class x_install_scripts_bin(x_install_scripts_custom):
    dir_name = "bin"
    var_name = "bindir"


class x_install_scripts_sbin(x_install_scripts_custom):
    dir_name = "sbin"
    var_name = "sbindir"


class x_install_scripts_portagebin(x_install_scripts_custom):
    dir_name = "portage"
    var_name = "portage_bindir"


class x_install_scripts(install_scripts):
    def initialize_option(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.run_command("install_scripts_bin")
        self.run_command("install_scripts_portagebin")
        self.run_command("install_scripts_sbin")


class x_sdist(sdist):
    """sdist defaulting to .tar.bz2 format, and archive files owned by root"""

    def finalize_options(self):
        if self.owner is None:
            self.owner = "root"
        if self.group is None:
            self.group = "root"

        sdist.finalize_options(self)


class build_tests(x_build_scripts_custom):
    """Prepare build dir for running tests."""

    def initialize_options(self):
        x_build_scripts_custom.initialize_options(self)
        self.build_base = None
        self.build_lib = None

    def finalize_options(self):
        x_build_scripts_custom.finalize_options(self)
        self.set_undefined_options(
            "build", ("build_base", "build_base"), ("build_lib", "build_lib")
        )

        # since we will be writing to $build_lib/.., it is important
        # that we do not leave $build_base
        self.top_dir = os.path.normpath(os.path.join(self.build_lib, ".."))
        cprefix = os.path.commonprefix((self.build_base, self.top_dir))
        if cprefix != self.build_base:
            raise SystemError("build_lib must be a subdirectory of build_base")

        self.build_dir = os.path.join(self.top_dir, "bin")

    def run(self):
        self.run_command("build_py")

        # install all scripts $build_lib/../bin
        # (we can't do a symlink since we want shebangs corrected)
        x_build_scripts_custom.run(self)

        # symlink 'cnf' directory
        conf_dir = os.path.join(self.top_dir, "cnf")
        if os.path.exists(conf_dir):
            if not os.path.islink(conf_dir):
                raise SystemError(
                    "%s exists and is not a symlink (collision)" % repr(conf_dir)
                )
            os.unlink(conf_dir)
        conf_src = os.path.relpath("cnf", self.top_dir)
        print("Symlinking %s -> %s" % (conf_dir, conf_src))
        os.symlink(conf_src, conf_dir)

        source_path = os.path.realpath(__file__)

        # copy GPG test keys
        copy_tree(
            os.path.join(
                os.path.dirname(source_path), "lib", "portage", "tests", ".gnupg"
            ),
            os.path.join(self.build_lib, "portage", "tests", ".gnupg"),
        )
        os.chmod(os.path.join(self.build_lib, "portage", "tests", ".gnupg"), 0o700)

        # create $build_lib/../.portage_not_installed
        # to enable proper paths in tests
        with open(os.path.join(self.top_dir, ".portage_not_installed"), "w"):
            pass


class test(Command):
    """run tests"""

    user_options = []

    def initialize_options(self):
        self.build_lib = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_lib", "build_lib"))

    def run(self):
        self.run_command("build_tests")

        subprocess.check_call(
            [
                sys.executable,
                "-bWd",
                os.path.join(self.build_lib, "portage/tests/runTests.py"),
            ]
        )


def find_packages():
    for dirpath, _dirnames, filenames in os.walk("lib"):
        if "__init__.py" in filenames:
            yield os.path.relpath(dirpath, "lib")


def find_scripts():
    for dirpath, _dirnames, filenames in os.walk("bin"):
        for f in filenames:
            if f not in ["deprecated-path"]:
                yield os.path.join(dirpath, f)


def get_manpages():
    linguas = os.environ.get("LINGUAS")
    if linguas is not None:
        linguas = linguas.split()

    for dirpath, _dirnames, filenames in os.walk("man"):
        groups = collections.defaultdict(list)
        for f in filenames:
            _fn, suffix = f.rsplit(".", 1)
            groups[suffix].append(os.path.join(dirpath, f))

        topdir = dirpath[len("man/") :]
        if not topdir or linguas is None or topdir in linguas:
            for g, mans in groups.items():
                yield [os.path.join("$mandir", topdir, "man%s" % g), mans]


class build_ext(_build_ext):
    user_options = _build_ext.user_options + [
        (
            "portage-ext-modules",
            None,
            "enable portage's C/C++ extensions (cross-compiling is not supported)",
        ),
    ]

    boolean_options = _build_ext.boolean_options + [
        "portage_ext_modules",
    ]

    def initialize_options(self):
        _build_ext.initialize_options(self)
        self.portage_ext_modules = None

    def run(self):
        if self.portage_ext_modules:
            _build_ext.run(self)


def venv_data_files(locations):
    if not create_entry_points:
        return
    for dest_prefix, source_path, file_args in locations:
        specific_files = []
        mode_arg = None
        for arg in file_args:
            if arg.startswith("-m"):
                mode_arg = int(arg[2:], 8)
            else:
                specific_files.append(arg)

        abs_source_path = os.path.abspath(source_path)
        for root, dirs, files in os.walk(abs_source_path):

            root_offset = root[len(abs_source_path) :].lstrip("/")
            dest_path = os.path.join(dest_prefix, root_offset)

            if specific_files:
                matched_files = list(
                    itertools.chain.from_iterable(
                        glob.glob(os.path.join(root, x)) for x in specific_files
                    )
                )
            else:
                matched_files = [os.path.join(root, x) for x in files]

            if mode_arg:
                for filename in matched_files:
                    if not os.path.islink(filename):
                        os.chmod(filename, mode_arg)

            yield (dest_path, matched_files)


def get_data_files(regular_files, venv_files):
    if create_entry_points:
        return list(venv_data_files(venv_files))

    return regular_files


setup(
    name="portage",
    version="3.0.36",
    url="https://wiki.gentoo.org/wiki/Project:Portage",
    project_urls={
        "Release Notes": "https://gitweb.gentoo.org/proj/portage.git/plain/RELEASE-NOTES",
        "Documentation": "https://wiki.gentoo.org/wiki/Handbook:AMD64/Working/Portage",
    },
    author="Gentoo Portage Development Team",
    author_email="dev-portage@gentoo.org",
    description="Portage is the package management and distribution system for Gentoo",
    license="GPLV2",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"": "lib"},
    packages=list(find_packages()),
    # something to cheat build & install commands
    scripts=list(find_scripts()),
    data_files=get_data_files(
        list(get_manpages())
        + [
            ["$sysconfdir", ["cnf/etc-update.conf", "cnf/dispatch-conf.conf"]],
            ["$logrotatedir", ["cnf/logrotate.d/elog-save-summary"]],
            [
                "$portage_confdir",
                ["cnf/make.conf.example", "cnf/make.globals", "cnf/repos.conf"],
            ],
            ["$portage_setsdir", ["cnf/sets/portage.conf"]],
            ["$docdir", ["NEWS", "RELEASE-NOTES"]],
            ["$portage_base/bin", ["bin/deprecated-path"]],
            ["$portage_confdir/repo.postsync.d", ["cnf/repo.postsync.d/example"]],
        ],
        [
            ("etc", "cnf", ("etc-update.conf", "dispatch-conf.conf")),
            ("etc/logrotate.d", "cnf/logrotate.d", ("elog-save-summary",)),
            (
                "share/portage/config/repo.postsync.d",
                "cnf/repo.postsync.d",
                ("example",),
            ),
            (
                "share/portage/config",
                "cnf",
                ("make.conf.example", "make.globals", "repos.conf"),
            ),
            ("share/portage/config/sets", "cnf/sets", ("*.conf",)),
            ("share/man/man1", "man", ("*.1",)),
            ("share/man/man5", "man", ("*.5",)),
            ("share/portage/doc", "", ("NEWS", "RELEASE-NOTES")),
            ("lib/portage/bin", "bin", ("-m0755",)),
        ],
    ),
    entry_points={
        "console_scripts": [
            "{}=portage.util.bin_entry_point:bin_entry_point".format(
                os.path.basename(path)
            )
            for path in itertools.chain.from_iterable(x_scripts.values())
        ],
    }
    if create_entry_points
    else {},
    # create_entry_points disables ext_modules, for pure python
    ext_modules=[]
    if create_entry_points
    else [
        Extension(
            name=n,
            sources=m,
            extra_compile_args=[
                "-D_FILE_OFFSET_BITS=64",
                "-D_LARGEFILE_SOURCE",
                "-D_LARGEFILE64_SOURCE",
            ],
        )
        for n, m in x_c_helpers.items()
    ],
    cmdclass={
        "build": x_build,
        "build_ext": build_ext,
        "build_man": build_man,
        "build_scripts": x_build_scripts,
        "build_scripts_bin": x_build_scripts_bin,
        "build_scripts_portagebin": x_build_scripts_portagebin,
        "build_scripts_sbin": x_build_scripts_sbin,
        "build_tests": build_tests,
        "clean": x_clean,
        "docbook": docbook,
        "apidoc": apidoc,
        "install": x_install,
        "install_data": x_install_data,
        "install_docbook": install_docbook,
        "install_apidoc": install_apidoc,
        "install_lib": x_install_lib,
        "install_scripts": x_install_scripts,
        "install_scripts_bin": x_install_scripts_bin,
        "install_scripts_portagebin": x_install_scripts_portagebin,
        "install_scripts_sbin": x_install_scripts_sbin,
        "sdist": x_sdist,
        "test": test,
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Installation/Setup",
    ],
    python_requires=">=3.6",
)



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import itertools


def dnf_convert(dep_struct):
    """
    Convert dep_struct to disjunctive normal form (DNF), where dep_struct
    is either a conjunction or disjunction of the form produced by
    use_reduce(opconvert=True).
    """
    # Normalize input to have a top-level conjunction.
    if isinstance(dep_struct, list):
        if dep_struct and dep_struct[0] == "||":
            dep_struct = [dep_struct]
    else:
        dep_struct = [dep_struct]

    conjunction = []
    disjunctions = []

    for x in dep_struct:
        if isinstance(x, list):
            assert (
                x and x[0] == "||"
            ), "Normalization error, nested conjunction found in %s" % (dep_struct,)
            if any(isinstance(element, list) for element in x):
                x_dnf = ["||"]
                for element in x[1:]:
                    if isinstance(element, list):
                        # Due to normalization, a disjunction must not be
                        # nested directly in another disjunction, so this
                        # must be a conjunction.
                        assert (
                            element
                        ), "Normalization error, empty conjunction found in %s" % (x,)
                        assert (
                            element[0] != "||"
                        ), "Normalization error, nested disjunction found in %s" % (x,)
                        element = dnf_convert(element)
                        if contains_disjunction(element):
                            assert (
                                len(element) == 1
                                and element[0]
                                and element[0][0] == "||"
                            ), (
                                "Normalization error, expected single disjunction in %s"
                                % (element,)
                            )
                            x_dnf.extend(element[0][1:])
                        else:
                            x_dnf.append(element)
                    else:
                        x_dnf.append(element)
                x = x_dnf
            disjunctions.append(x)
        else:
            conjunction.append(x)

    if disjunctions and (conjunction or len(disjunctions) > 1):
        dnf_form = ["||"]
        for x in itertools.product(*[x[1:] for x in disjunctions]):
            normalized = conjunction[:]
            for element in x:
                if isinstance(element, list):
                    normalized.extend(element)
                else:
                    normalized.append(element)
            dnf_form.append(normalized)
        result = [dnf_form]
    else:
        result = conjunction + disjunctions

    return result


def contains_disjunction(dep_struct):
    """
    Search for a disjunction contained in dep_struct, where dep_struct
    is either a conjunction or disjunction of the form produced by
    use_reduce(opconvert=True). If dep_struct is a disjunction, then
    this only returns True if there is a nested disjunction. Due to
    normalization, recursion is only needed when dep_struct is a
    disjunction containing a conjunction. If dep_struct is a conjunction,
    then it is assumed that normalization has elevated any nested
    disjunctions to the top-level.
    """
    is_disjunction = dep_struct and dep_struct[0] == "||"
    for x in dep_struct:
        if isinstance(x, list):
            assert x, "Normalization error, empty conjunction found in %s" % (
                dep_struct,
            )
            if x[0] == "||":
                return True
            if is_disjunction and contains_disjunction(x):
                return True
    return False



# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""Contains private support functions for the Display class
in output.py
"""

__all__ = ()

import re

from portage import os
from portage._sets.base import InternalPackageSet
from portage.exception import PackageSetNotFound
from portage.localization import localized_size
from portage.output import (
    blue,
    colorize,
    create_color_func,
    green,
    red,
    teal,
    turquoise,
    yellow,
)

bad = create_color_func("BAD")
from portage.util import writemsg
from portage.util.SlotObject import SlotObject

from _emerge.Blocker import Blocker
from _emerge.Package import Package


class _RepoDisplay:
    def __init__(self, roots):
        self._shown_repos = {}
        self._unknown_repo = False
        repo_paths = set()
        for root_config in roots.values():
            for repo in root_config.settings.repositories:
                repo_paths.add(repo.location)
        repo_paths = list(repo_paths)
        self._repo_paths = repo_paths
        self._repo_paths_real = [
            os.path.realpath(repo_path) for repo_path in repo_paths
        ]

    def repoStr(self, repo_path_real):
        real_index = -1
        if repo_path_real:
            real_index = self._repo_paths_real.index(repo_path_real)
        if real_index == -1:
            s = "?"
            self._unknown_repo = True
        else:
            shown_repos = self._shown_repos
            repo_paths = self._repo_paths
            repo_path = repo_paths[real_index]
            index = shown_repos.get(repo_path)
            if index is None:
                index = len(shown_repos)
                shown_repos[repo_path] = index
            s = str(index)
        return s

    def __str__(self):
        output = []
        shown_repos = self._shown_repos
        unknown_repo = self._unknown_repo
        if shown_repos or self._unknown_repo:
            output.append("Repositories:\n")
        show_repo_paths = list(shown_repos)
        for repo_path, repo_index in shown_repos.items():
            show_repo_paths[repo_index] = repo_path
        if show_repo_paths:
            for index, repo_path in enumerate(show_repo_paths):
                output.append(" " + teal("[" + str(index) + "]") + " %s\n" % repo_path)
        if unknown_repo:
            output.append(
                " "
                + teal("[?]")
                + " indicates that the source repository could not be determined\n"
            )
        return "".join(output)


class _PackageCounters:
    def __init__(self):
        self.upgrades = 0
        self.downgrades = 0
        self.new = 0
        self.newslot = 0
        self.reinst = 0
        self.uninst = 0
        self.blocks = 0
        self.blocks_satisfied = 0
        self.totalsize = 0
        self.restrict_fetch = 0
        self.restrict_fetch_satisfied = 0
        self.interactive = 0
        self.binary = 0

    def __str__(self):
        total_installs = (
            self.upgrades + self.downgrades + self.newslot + self.new + self.reinst
        )
        myoutput = []
        details = []
        myoutput.append("Total: %s package" % total_installs)
        if total_installs != 1:
            myoutput.append("s")
        if total_installs != 0:
            myoutput.append(" (")
        if self.upgrades > 0:
            details.append("%s upgrade" % self.upgrades)
            if self.upgrades > 1:
                details[-1] += "s"
        if self.downgrades > 0:
            details.append("%s downgrade" % self.downgrades)
            if self.downgrades > 1:
                details[-1] += "s"
        if self.new > 0:
            details.append("%s new" % self.new)
        if self.newslot > 0:
            details.append("%s in new slot" % self.newslot)
            if self.newslot > 1:
                details[-1] += "s"
        if self.reinst > 0:
            details.append("%s reinstall" % self.reinst)
            if self.reinst > 1:
                details[-1] += "s"
        if self.binary > 0:
            details.append("%s binary" % self.binary)
            if self.binary > 1:
                details[-1] = details[-1][:-1] + "ies"
        if self.uninst > 0:
            details.append("%s uninstall" % self.uninst)
            if self.uninst > 1:
                details[-1] += "s"
        if self.interactive > 0:
            details.append(
                "%s %s" % (self.interactive, colorize("WARN", "interactive"))
            )
        myoutput.append(", ".join(details))
        if total_installs != 0:
            myoutput.append(")")
        myoutput.append(", Size of downloads: %s" % localized_size(self.totalsize))
        if self.restrict_fetch:
            myoutput.append("\nFetch Restriction: %s package" % self.restrict_fetch)
            if self.restrict_fetch > 1:
                myoutput.append("s")
        if self.restrict_fetch_satisfied < self.restrict_fetch:
            myoutput.append(
                bad(" (%s unsatisfied)")
                % (self.restrict_fetch - self.restrict_fetch_satisfied)
            )
        if self.blocks > 0:
            myoutput.append("\nConflict: %s block" % self.blocks)
            if self.blocks > 1:
                myoutput.append("s")
            if self.blocks_satisfied < self.blocks:
                myoutput.append(
                    bad(" (%s unsatisfied)") % (self.blocks - self.blocks_satisfied)
                )
            else:
                myoutput.append(" (all satisfied)")
        return "".join(myoutput)


class _DisplayConfig:
    def __init__(self, depgraph, mylist, favorites, verbosity):
        frozen_config = depgraph._frozen_config
        dynamic_config = depgraph._dynamic_config

        self.mylist = mylist
        self.favorites = InternalPackageSet(favorites, allow_repo=True)
        self.verbosity = verbosity

        if self.verbosity is None:
            self.verbosity = (
                "--quiet" in frozen_config.myopts
                and 1
                or "--verbose" in frozen_config.myopts
                and 3
                or 2
            )

        self.oneshot = (
            "--oneshot" in frozen_config.myopts or "--onlydeps" in frozen_config.myopts
        )
        self.columns = "--columns" in frozen_config.myopts
        self.tree_display = "--tree" in frozen_config.myopts
        self.alphabetical = "--alphabetical" in frozen_config.myopts
        self.quiet = "--quiet" in frozen_config.myopts
        self.all_flags = self.verbosity == 3 or self.quiet
        self.print_use_string = (
            self.verbosity != 1 or "--verbose" in frozen_config.myopts
        )
        self.edebug = frozen_config.edebug
        self.unordered_display = "--unordered-display" in frozen_config.myopts

        mywidth = 130
        if "COLUMNWIDTH" in frozen_config.settings:
            try:
                mywidth = int(frozen_config.settings["COLUMNWIDTH"])
            except ValueError as e:
                writemsg("!!! %s\n" % str(e), noiselevel=-1)
                writemsg(
                    "!!! Unable to parse COLUMNWIDTH='%s'\n"
                    % frozen_config.settings["COLUMNWIDTH"],
                    noiselevel=-1,
                )
                del e
        self.columnwidth = mywidth

        if "--quiet-repo-display" in frozen_config.myopts:
            self.repo_display = _RepoDisplay(frozen_config.roots)
        self.trees = frozen_config.trees
        self.pkgsettings = frozen_config.pkgsettings
        self.target_root = frozen_config.target_root
        self.running_root = frozen_config._running_root
        self.roots = frozen_config.roots

        # Create a set of selected packages for each root
        self.selected_sets = {}
        for root_name, root in self.roots.items():
            try:
                self.selected_sets[root_name] = InternalPackageSet(
                    initial_atoms=root.setconfig.getSetAtoms("selected")
                )
            except PackageSetNotFound:
                # A nested set could not be resolved, so ignore nested sets.
                self.selected_sets[root_name] = root.sets["selected"]

        self.blocker_parents = dynamic_config._blocker_parents
        self.reinstall_nodes = dynamic_config._reinstall_nodes
        self.digraph = dynamic_config.digraph
        self.blocker_uninstalls = dynamic_config._blocker_uninstalls
        self.package_tracker = dynamic_config._package_tracker
        self.set_nodes = dynamic_config._set_nodes

        self.pkg_use_enabled = depgraph._pkg_use_enabled
        self.pkg = depgraph._pkg


_alnum_sort_re = re.compile(r"(\d+)")


def _alnum_sort_key(x):
    def _convert_even_to_int(it):
        it = iter(it)
        try:
            while True:
                yield next(it)
                yield int(next(it))
        except StopIteration:
            pass

    return tuple(_convert_even_to_int(_alnum_sort_re.split(x)))


def _create_use_string(
    conf,
    name,
    cur_iuse,
    iuse_forced,
    cur_use,
    old_iuse,
    old_use,
    is_new,
    feature_flags,
    reinst_flags,
):

    if not conf.print_use_string:
        return ""

    enabled = []
    if conf.alphabetical:
        disabled = enabled
        removed = enabled
    else:
        disabled = []
        removed = []
    cur_iuse = set(cur_iuse)
    enabled_flags = cur_iuse.intersection(cur_use)
    removed_iuse = set(old_iuse).difference(cur_iuse)
    any_iuse = cur_iuse.union(old_iuse)
    any_iuse = list(any_iuse)
    any_iuse.sort(key=_alnum_sort_key)

    for flag in any_iuse:
        flag_str = None
        isEnabled = False
        reinst_flag = reinst_flags and flag in reinst_flags
        if flag in enabled_flags:
            isEnabled = True
            if is_new or flag in old_use and (conf.all_flags or reinst_flag):
                flag_str = red(flag)
            elif flag not in old_iuse:
                flag_str = yellow(flag) + "%*"
            elif flag not in old_use:
                flag_str = green(flag) + "*"
        elif flag in removed_iuse:
            if conf.all_flags or reinst_flag:
                flag_str = yellow("-" + flag) + "%"
                if flag in old_use:
                    flag_str += "*"
                flag_str = "(" + flag_str + ")"
                removed.append(flag_str)
            continue
        else:
            if (
                is_new
                or flag in old_iuse
                and flag not in old_use
                and (conf.all_flags or reinst_flag)
            ):
                flag_str = blue("-" + flag)
            elif flag not in old_iuse:
                flag_str = yellow("-" + flag)
                if flag not in iuse_forced:
                    flag_str += "%"
            elif flag in old_use:
                flag_str = green("-" + flag) + "*"
        if flag_str:
            if flag in feature_flags:
                flag_str = "{" + flag_str + "}"
            elif flag in iuse_forced:
                flag_str = "(" + flag_str + ")"
            if isEnabled:
                enabled.append(flag_str)
            else:
                disabled.append(flag_str)

    if conf.alphabetical:
        ret = " ".join(enabled)
    else:
        ret = " ".join(enabled + disabled + removed)
    if ret:
        ret = '%s="%s" ' % (name, ret)
    return ret


def _tree_display(conf, mylist):

    # If there are any Uninstall instances, add the
    # corresponding blockers to the digraph.
    mygraph = conf.digraph.copy()

    executed_uninstalls = set(
        node
        for node in mylist
        if isinstance(node, Package) and node.operation == "unmerge"
    )

    for uninstall in conf.blocker_uninstalls.leaf_nodes():
        uninstall_parents = conf.blocker_uninstalls.parent_nodes(uninstall)
        if not uninstall_parents:
            continue

        # Remove the corresponding "nomerge" node and substitute
        # the Uninstall node.
        inst_pkg = conf.pkg(
            uninstall.cpv, "installed", uninstall.root_config, installed=True
        )

        try:
            mygraph.remove(inst_pkg)
        except KeyError:
            pass

        try:
            inst_pkg_blockers = conf.blocker_parents.child_nodes(inst_pkg)
        except KeyError:
            inst_pkg_blockers = []

        # Break the Package -> Uninstall edges.
        mygraph.remove(uninstall)

        # Resolution of a package's blockers
        # depend on it's own uninstallation.
        for blocker in inst_pkg_blockers:
            mygraph.add(uninstall, blocker)

        # Expand Package -> Uninstall edges into
        # Package -> Blocker -> Uninstall edges.
        for blocker in uninstall_parents:
            mygraph.add(uninstall, blocker)
            for parent in conf.blocker_parents.parent_nodes(blocker):
                if parent != inst_pkg:
                    mygraph.add(blocker, parent)

        # If the uninstall task did not need to be executed because
        # of an upgrade, display Blocker -> Upgrade edges since the
        # corresponding Blocker -> Uninstall edges will not be shown.
        upgrade_node = next(
            conf.package_tracker.match(uninstall.root, uninstall.slot_atom), None
        )

        if upgrade_node is not None and uninstall not in executed_uninstalls:
            for blocker in uninstall_parents:
                mygraph.add(upgrade_node, blocker)

    if conf.unordered_display:
        display_list = _unordered_tree_display(mygraph, mylist)
    else:
        display_list = _ordered_tree_display(conf, mygraph, mylist)

    _prune_tree_display(display_list)

    return display_list


def _unordered_tree_display(mygraph, mylist):
    display_list = []
    seen_nodes = set()

    def print_node(node, depth):

        if node in seen_nodes:
            pass
        else:
            seen_nodes.add(node)

            if isinstance(node, (Blocker, Package)):
                display_list.append((node, depth, True))
            else:
                depth = -1

            for child_node in mygraph.child_nodes(node):
                print_node(child_node, depth + 1)

    for root_node in mygraph.root_nodes():
        print_node(root_node, 0)

    return display_list


def _ordered_tree_display(conf, mygraph, mylist):
    depth = 0
    shown_edges = set()
    tree_nodes = []
    display_list = []

    for x in mylist:
        depth = len(tree_nodes)
        while depth and x not in mygraph.child_nodes(tree_nodes[depth - 1]):
            depth -= 1
        if depth:
            tree_nodes = tree_nodes[:depth]
            tree_nodes.append(x)
            display_list.append((x, depth, True))
            shown_edges.add((x, tree_nodes[depth - 1]))
        else:
            traversed_nodes = set()  # prevent endless circles
            traversed_nodes.add(x)

            def add_parents(current_node, ordered):
                parent_nodes = None
                # Do not traverse to parents if this node is an
                # an argument or a direct member of a set that has
                # been specified as an argument (system or world).
                if current_node not in conf.set_nodes:
                    parent_nodes = mygraph.parent_nodes(current_node)
                if parent_nodes:
                    child_nodes = set(mygraph.child_nodes(current_node))
                    selected_parent = None
                    # First, try to avoid a direct cycle.
                    for node in parent_nodes:
                        if not isinstance(node, (Blocker, Package)):
                            continue
                        if node not in traversed_nodes and node not in child_nodes:
                            edge = (current_node, node)
                            if edge in shown_edges:
                                continue
                            selected_parent = node
                            break
                    if not selected_parent:
                        # A direct cycle is unavoidable.
                        for node in parent_nodes:
                            if not isinstance(node, (Blocker, Package)):
                                continue
                            if node not in traversed_nodes:
                                edge = (current_node, node)
                                if edge in shown_edges:
                                    continue
                                selected_parent = node
                                break
                    if selected_parent:
                        shown_edges.add((current_node, selected_parent))
                        traversed_nodes.add(selected_parent)
                        add_parents(selected_parent, False)
                display_list.append((current_node, len(tree_nodes), ordered))
                tree_nodes.append(current_node)

            tree_nodes = []
            add_parents(x, True)

    return display_list


def _prune_tree_display(display_list):
    last_merge_depth = 0
    for i in range(len(display_list) - 1, -1, -1):
        node, depth, ordered = display_list[i]
        if (
            not ordered
            and depth == 0
            and i > 0
            and node == display_list[i - 1][0]
            and display_list[i - 1][1] == 0
        ):
            # An ordered node got a consecutive duplicate
            # when the tree was being filled in.
            del display_list[i]
            continue
        if (
            ordered
            and isinstance(node, Package)
            and node.operation in ("merge", "uninstall")
        ):
            last_merge_depth = depth
            continue
        if (
            depth >= last_merge_depth
            or i < len(display_list) - 1
            and depth >= display_list[i + 1][1]
        ):
            del display_list[i]


def _strip_header_comments(lines):
    # strip leading and trailing blank or header/comment lines
    i = 0
    while i < len(lines) and (not lines[i] or lines[i][:1] == "#"):
        i += 1
    if i:
        lines = lines[i:]
    while lines and (not lines[-1] or lines[-1][:1] == "#"):
        lines.pop()
    return lines


class PkgInfo:
    """Simple class to hold instance attributes for current
    information about the pkg being printed.
    """

    __slots__ = (
        "attr_display",
        "built",
        "cp",
        "ebuild_path",
        "fetch_symbol",
        "merge",
        "oldbest",
        "oldbest_list",
        "operation",
        "ordered",
        "previous_pkg",
        "repo_name",
        "repo_path_real",
        "slot",
        "sub_slot",
        "system",
        "use",
        "ver",
        "world",
    )

    def __init__(self):
        self.built = False
        self.cp = ""
        self.ebuild_path = ""
        self.fetch_symbol = ""
        self.merge = ""
        self.oldbest = ""
        self.oldbest_list = []
        self.operation = ""
        self.ordered = False
        self.previous_pkg = None
        self.repo_path_real = ""
        self.repo_name = ""
        self.slot = ""
        self.sub_slot = ""
        self.system = False
        self.use = ""
        self.ver = ""
        self.world = False
        self.attr_display = PkgAttrDisplay()


class PkgAttrDisplay(SlotObject):

    __slots__ = (
        "downgrade",
        "fetch_restrict",
        "fetch_restrict_satisfied",
        "force_reinstall",
        "interactive",
        "mask",
        "new",
        "new_slot",
        "new_version",
        "replace",
    )

    def __str__(self):
        output = []

        if self.interactive:
            output.append(colorize("WARN", "I"))
        else:
            output.append(" ")

        if self.new or self.force_reinstall:
            if self.force_reinstall:
                output.append(red("r"))
            else:
                output.append(green("N"))
        else:
            output.append(" ")

        if self.new_slot or self.replace:
            if self.replace:
                output.append(yellow("R"))
            else:
                output.append(green("S"))
        else:
            output.append(" ")

        if self.fetch_restrict or self.fetch_restrict_satisfied:
            if self.fetch_restrict_satisfied:
                output.append(green("f"))
            else:
                output.append(red("F"))
        else:
            output.append(" ")

        if self.new_version:
            output.append(turquoise("U"))
        else:
            output.append(" ")

        if self.downgrade:
            output.append(blue("D"))
        else:
            output.append(" ")

        if self.mask is not None:
            output.append(self.mask)

        return "".join(output)



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "AbstractChildWatcher",
    "DefaultEventLoopPolicy",
)

import asyncio as _real_asyncio
from asyncio import events
from asyncio.unix_events import AbstractChildWatcher

import fcntl
import os

from portage.util._eventloop.global_event_loop import (
    global_event_loop as _global_event_loop,
)


if hasattr(os, "set_blocking"):

    def _set_nonblocking(fd):
        os.set_blocking(fd, False)

else:

    def _set_nonblocking(fd):
        flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        flags = flags | os.O_NONBLOCK
        fcntl.fcntl(fd, fcntl.F_SETFL, flags)


class _PortageEventLoopPolicy(events.AbstractEventLoopPolicy):
    """
    Implementation of asyncio.AbstractEventLoopPolicy based on portage's
    internal event loop. This supports running event loops in forks,
    which is not supported by the default asyncio event loop policy,
    see https://bugs.python.org/issue22087.
    """

    def get_event_loop(self):
        """
        Get the event loop for the current context.

        Returns an event loop object implementing the AbstractEventLoop
        interface.

        @rtype: asyncio.AbstractEventLoop (or compatible)
        @return: the current event loop policy
        """
        return _global_event_loop()._asyncio_wrapper

    def get_child_watcher(self):
        """Get the watcher for child processes."""
        return _global_event_loop()._asyncio_child_watcher


class _AsyncioEventLoopPolicy(_PortageEventLoopPolicy):
    """
    A subclass of _PortageEventLoopPolicy which raises
    NotImplementedError if it is set as the real asyncio event loop
    policy, since this class is intended to *wrap* the real asyncio
    event loop policy.
    """

    def _check_recursion(self):
        if _real_asyncio.get_event_loop_policy() is self:
            raise NotImplementedError("this class is only a wrapper")

    def get_event_loop(self):
        self._check_recursion()
        return super(_AsyncioEventLoopPolicy, self).get_event_loop()

    def get_child_watcher(self):
        self._check_recursion()
        return super(_AsyncioEventLoopPolicy, self).get_child_watcher()


DefaultEventLoopPolicy = _AsyncioEventLoopPolicy



# Copyright 1998-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__docformat__ = "epytext"

import errno
import io
import itertools
import re
import subprocess
import sys

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.util:writemsg",
)
import portage.util.formatter as formatter

from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage import _unicode_decode
from portage.const import COLOR_MAP_FILE
from portage.exception import (
    CommandNotFound,
    FileNotFound,
    ParseError,
    PermissionDenied,
    PortageException,
)
from portage.localization import _

havecolor = 1
dotitles = 1

"""Maps style class to tuple of attribute names."""
_styles = {}

"""Maps attribute name to ansi code."""

esc_seq = "\x1b["

codes = {
    "normal": f"{esc_seq}0m",
    "reset": f"{esc_seq}39;49;00m",
    "bold": f"{esc_seq}01m",
    "faint": f"{esc_seq}02m",
    "standout": f"{esc_seq}03m",
    "underline": f"{esc_seq}04m",
    "blink": f"{esc_seq}05m",
    "overline": f"{esc_seq}06m",
    "reverse": f"{esc_seq}07m",
    "invisible": f"{esc_seq}08m",
    "no-attr": f"{esc_seq}22m",
    "no-standout": f"{esc_seq}23m",
    "no-underline": f"{esc_seq}24m",
    "no-blink": f"{esc_seq}25m",
    "no-overline": f"{esc_seq}26m",
    "no-reverse": f"{esc_seq}27m",
    "bg_black": f"{esc_seq}40m",
    "bg_darkred": f"{esc_seq}41m",
    "bg_darkgreen": f"{esc_seq}42m",
    "bg_brown": f"{esc_seq}43m",
    "bg_darkblue": f"{esc_seq}44m",
    "bg_purple": f"{esc_seq}45m",
    "bg_teal": f"{esc_seq}46m",
    "bg_lightgray": f"{esc_seq}47m",
    "bg_default": f"{esc_seq}49m",
    "bg_darkyellow": f"{esc_seq}43m",
}


def color(fg, bg="default", attr=["normal"]):
    myansicodechain = itertools.chain((codes[fg]), (codes[x] for x in [bg, *attr]))
    return "".join(myansicodechain)


ansi_codes = [y for x in range(30, 38) for y in (f"{x}m", f"{x};01m")]


rgb_ansi_colors = [
    "0x000000",
    "0x555555",
    "0xAA0000",
    "0xFF5555",
    "0x00AA00",
    "0x55FF55",
    "0xAA5500",
    "0xFFFF55",
    "0x0000AA",
    "0x5555FF",
    "0xAA00AA",
    "0xFF55FF",
    "0x00AAAA",
    "0x55FFFF",
    "0xAAAAAA",
    "0xFFFFFF",
]

for x in range(len(rgb_ansi_colors)):
    codes[rgb_ansi_colors[x]] = esc_seq + ansi_codes[x]

del x

codes["black"] = codes["0x000000"]
codes["darkgray"] = codes["0x555555"]

codes["red"] = codes["0xFF5555"]
codes["darkred"] = codes["0xAA0000"]

codes["green"] = codes["0x55FF55"]
codes["darkgreen"] = codes["0x00AA00"]

codes["yellow"] = codes["0xFFFF55"]
codes["brown"] = codes["0xAA5500"]

codes["blue"] = codes["0x5555FF"]
codes["darkblue"] = codes["0x0000AA"]

codes["fuchsia"] = codes["0xFF55FF"]
codes["purple"] = codes["0xAA00AA"]

codes["turquoise"] = codes["0x55FFFF"]
codes["teal"] = codes["0x00AAAA"]

codes["white"] = codes["0xFFFFFF"]
codes["lightgray"] = codes["0xAAAAAA"]

codes["darkteal"] = codes["turquoise"]
# Some terminals have darkyellow instead of brown.
codes["0xAAAA00"] = codes["brown"]
codes["darkyellow"] = codes["0xAAAA00"]


# Colors from /lib/gentoo/functions.sh
_styles["BAD"] = ("red",)
_styles["BRACKET"] = ("blue",)
_styles["ERR"] = ("red",)
_styles["GOOD"] = ("green",)
_styles["HILITE"] = ("teal",)
_styles["INFO"] = ("darkgreen",)
_styles["LOG"] = ("green",)
_styles["NORMAL"] = ("normal",)
_styles["QAWARN"] = ("brown",)
_styles["WARN"] = ("yellow",)

# Portage functions
_styles["INFORM"] = ("darkgreen",)
_styles["UNMERGE_WARN"] = ("red",)
_styles["SECURITY_WARN"] = ("red",)
_styles["MERGE_LIST_PROGRESS"] = ("yellow",)
_styles["PKG_BLOCKER"] = ("red",)
_styles["PKG_BLOCKER_SATISFIED"] = ("teal",)
_styles["PKG_MERGE"] = ("darkgreen",)
_styles["PKG_MERGE_SYSTEM"] = ("darkgreen",)
_styles["PKG_MERGE_WORLD"] = ("green",)
_styles["PKG_BINARY_MERGE"] = ("purple",)
_styles["PKG_BINARY_MERGE_SYSTEM"] = ("purple",)
_styles["PKG_BINARY_MERGE_WORLD"] = ("fuchsia",)
_styles["PKG_UNINSTALL"] = ("red",)
_styles["PKG_NOMERGE"] = ("teal",)
_styles["PKG_NOMERGE_SYSTEM"] = ("teal",)
_styles["PKG_NOMERGE_WORLD"] = ("blue",)
_styles["PROMPT_CHOICE_DEFAULT"] = ("green",)
_styles["PROMPT_CHOICE_OTHER"] = ("red",)


def _parse_color_map(config_root="/", onerror=None):
    """
    Parse /etc/portage/color.map and return a dict of error codes.

    @param onerror: an optional callback to handle any ParseError that would
            otherwise be raised
    @type onerror: callable
    @rtype: dict
    @return: a dictionary mapping color classes to color codes
    """
    global codes, _styles
    myfile = os.path.join(config_root, COLOR_MAP_FILE)
    ansi_code_pattern = re.compile("^[0-9;]*m$")
    quotes = "'\""

    def strip_quotes(token):
        if token[0] in quotes and token[0] == token[-1]:
            token = token[1:-1]
        return token

    try:
        with io.open(
            _unicode_encode(myfile, encoding=_encodings["fs"], errors="strict"),
            mode="r",
            encoding=_encodings["content"],
            errors="replace",
        ) as f:
            lines = f.readlines()
        for lineno, line in enumerate(lines):
            commenter_pos = line.find("#")
            line = line[:commenter_pos].strip()

            if len(line) == 0:
                continue

            split_line = line.split("=")
            if len(split_line) != 2:
                e = ParseError(
                    _("'%s', line %s: expected exactly one occurrence of '=' operator")
                    % (myfile, lineno)
                )
                raise e
                if onerror:
                    onerror(e)
                else:
                    raise e
                continue

            k = strip_quotes(split_line[0].strip())
            v = strip_quotes(split_line[1].strip())
            if not k in _styles and not k in codes:
                e = ParseError(
                    _("'%s', line %s: Unknown variable: '%s'") % (myfile, lineno, k)
                )
                if onerror:
                    onerror(e)
                else:
                    raise e
                continue
            if ansi_code_pattern.match(v):
                if k in _styles:
                    _styles[k] = (esc_seq + v,)
                elif k in codes:
                    codes[k] = esc_seq + v
            else:
                code_list = []
                for x in v.split():
                    if x in codes:
                        if k in _styles:
                            code_list.append(x)
                        elif k in codes:
                            code_list.append(codes[x])
                    else:
                        e = ParseError(
                            _("'%s', line %s: Undefined: '%s'") % (myfile, lineno, x)
                        )
                        if onerror:
                            onerror(e)
                        else:
                            raise e
                if k in _styles:
                    _styles[k] = tuple(code_list)
                elif k in codes:
                    codes[k] = "".join(code_list)
    except (IOError, OSError) as e:
        if e.errno == errno.ENOENT:
            raise FileNotFound(myfile)
        elif e.errno == errno.EACCES:
            raise PermissionDenied(myfile)
        raise


def nc_len(mystr):
    tmp = re.sub(esc_seq + "^m]+m", "", mystr)
    return len(tmp)


_legal_terms_re = re.compile(
    r"^(xterm|xterm-color|Eterm|aterm|rxvt|screen|kterm|rxvt-unicode|gnome|interix|tmux|st-256color|alacritty|konsole|foot)"
)
_disable_xtermTitle = None
_max_xtermTitle_len = 253


def xtermTitle(mystr, raw=False):
    global _disable_xtermTitle
    if _disable_xtermTitle is None:
        _disable_xtermTitle = not (
            sys.__stderr__.isatty()
            and "TERM" in os.environ
            and _legal_terms_re.match(os.environ["TERM"]) is not None
        )

    if dotitles and not _disable_xtermTitle:
        # If the title string is too big then the terminal can
        # misbehave. Therefore, truncate it if it's too big.
        if len(mystr) > _max_xtermTitle_len:
            mystr = mystr[:_max_xtermTitle_len]
        if not raw:
            mystr = "\x1b]0;%s\x07" % mystr

        # avoid potential UnicodeEncodeError
        mystr = _unicode_encode(
            mystr, encoding=_encodings["stdio"], errors="backslashreplace"
        )
        f = sys.stderr.buffer
        f.write(mystr)
        f.flush()


default_xterm_title = None


def xtermTitleReset():
    global default_xterm_title
    if default_xterm_title is None:
        prompt_command = os.environ.get("PROMPT_COMMAND")
        if prompt_command == "":
            default_xterm_title = ""
        elif prompt_command is not None:
            if (
                dotitles
                and "TERM" in os.environ
                and _legal_terms_re.match(os.environ["TERM"]) is not None
                and sys.__stderr__.isatty()
            ):
                from portage.process import find_binary, spawn

                shell = os.environ.get("SHELL")
                if not shell or not os.access(shell, os.EX_OK):
                    shell = find_binary("sh")
                if shell:
                    spawn(
                        [shell, "-c", prompt_command],
                        env=os.environ,
                        fd_pipes={
                            0: portage._get_stdin().fileno(),
                            1: sys.__stderr__.fileno(),
                            2: sys.__stderr__.fileno(),
                        },
                    )
                else:
                    os.system(prompt_command)
            return
        else:
            pwd = os.environ.get("PWD", "")
            home = os.environ.get("HOME", "")
            if home != "" and pwd.startswith(home):
                pwd = "~" + pwd[len(home) :]
            default_xterm_title = "\x1b]0;%s@%s:%s\x07" % (
                os.environ.get("LOGNAME", ""),
                os.environ.get("HOSTNAME", "").split(".", 1)[0],
                pwd,
            )
    xtermTitle(default_xterm_title, raw=True)


def notitles():
    "turn off title setting"
    dotitles = 0


def nocolor():
    "turn off colorization"
    global havecolor
    havecolor = 0


def resetColor():
    return codes["reset"]


def style_to_ansi_code(style):
    """
    @param style: A style name
    @type style: String
    @rtype: String
    @return: A string containing one or more ansi escape codes that are
            used to render the given style.
    """
    ret = ""
    for attr_name in _styles[style]:
        # allow stuff that has found it's way through ansi_code_pattern
        ret += codes.get(attr_name, attr_name)
    return ret


def colormap():
    mycolors = []
    for c in (
        "BAD",
        "BRACKET",
        "ERR",
        "GOOD",
        "HILITE",
        "INFO",
        "LOG",
        "NORMAL",
        "QAWARN",
        "WARN",
    ):
        mycolors.append("PORTAGE_COLOR_{}=$'{}'".format(c, style_to_ansi_code(c)))
    return "\n".join(mycolors)


def colorize(color_key, text):
    global havecolor
    if havecolor:
        if color_key in codes:
            return codes[color_key] + text + codes["reset"]
        if color_key in _styles:
            return style_to_ansi_code(color_key) + text + codes["reset"]
    return text


compat_functions_colors = [
    "bold",
    "white",
    "teal",
    "turquoise",
    "darkteal",
    "fuchsia",
    "purple",
    "blue",
    "darkblue",
    "green",
    "darkgreen",
    "yellow",
    "brown",
    "darkyellow",
    "red",
    "darkred",
]


class create_color_func:
    __slots__ = ("_color_key",)

    def __init__(self, color_key):
        self._color_key = color_key

    def __call__(self, text):
        return colorize(self._color_key, text)


for c in compat_functions_colors:
    globals()[c] = create_color_func(c)


class ConsoleStyleFile:
    """
    A file-like object that behaves something like
    the colorize() function. Style identifiers
    passed in via the new_styles() method will be used to
    apply console codes to output.
    """

    def __init__(self, f):
        self._file = f
        self._styles = None
        self.write_listener = None

    def new_styles(self, styles):
        self._styles = styles

    def write(self, s):
        # In python-2.6, DumbWriter.send_line_break() can write
        # non-unicode '\n' which fails with TypeError if self._file
        # is a text stream such as io.StringIO. Therefore, make sure
        # input is converted to unicode when necessary.
        s = _unicode_decode(s)
        global havecolor
        if havecolor and self._styles:
            styled_s = []
            for style in self._styles:
                styled_s.append(style_to_ansi_code(style))
            styled_s.append(s)
            styled_s.append(codes["reset"])
            self._write(self._file, "".join(styled_s))
        else:
            self._write(self._file, s)
        if self.write_listener:
            self._write(self.write_listener, s)

    def _write(self, f, s):
        # avoid potential UnicodeEncodeError
        if f in (sys.stdout, sys.stderr):
            s = _unicode_encode(
                s, encoding=_encodings["stdio"], errors="backslashreplace"
            )
            f = f.buffer
        f.write(s)

    def writelines(self, lines):
        for s in lines:
            self.write(s)

    def flush(self):
        self._file.flush()

    def close(self):
        self._file.close()


class StyleWriter(formatter.DumbWriter):
    """
    This is just a DumbWriter with a hook in the new_styles() method
    that passes a styles tuple as a single argument to a callable
    style_listener attribute.
    """

    def __init__(self, **kwargs):
        formatter.DumbWriter.__init__(self, **kwargs)
        self.style_listener = None

    def new_styles(self, styles):
        formatter.DumbWriter.new_styles(self, styles)
        if self.style_listener:
            self.style_listener(styles)


def get_term_size(fd=None):
    """
    Get the number of lines and columns of the tty that is connected to
    fd.  Returns a tuple of (lines, columns) or (0, 0) if an error
    occurs. The curses module is used if available, otherwise the output of
    `stty size` is parsed. The lines and columns values are guaranteed to be
    greater than or equal to zero, since a negative COLUMNS variable is
    known to prevent some commands from working (see bug #394091).
    """
    if fd is None:
        fd = sys.stdout
    if not hasattr(fd, "isatty") or not fd.isatty():
        return (0, 0)
    try:
        import curses

        try:
            curses.setupterm(term=os.environ.get("TERM", "unknown"), fd=fd.fileno())
            return curses.tigetnum("lines"), curses.tigetnum("cols")
        except curses.error:
            pass
    except ImportError:
        pass

    try:
        proc = subprocess.Popen(["stty", "size"], stdout=subprocess.PIPE, stderr=fd)
    except EnvironmentError as e:
        if e.errno != errno.ENOENT:
            raise
        # stty command not found
        return (0, 0)

    out = _unicode_decode(proc.communicate()[0])
    if proc.wait() == os.EX_OK:
        out = out.split()
        if len(out) == 2:
            try:
                val = (int(out[0]), int(out[1]))
            except ValueError:
                pass
            else:
                if val[0] >= 0 and val[1] >= 0:
                    return val
    return (0, 0)


def set_term_size(lines, columns, fd):
    """
    Set the number of lines and columns for the tty that is connected to fd.
    For portability, this simply calls `stty rows $lines columns $columns`.
    """
    from portage.process import spawn

    cmd = ["stty", "rows", str(lines), "columns", str(columns)]
    try:
        spawn(cmd, env=os.environ, fd_pipes={0: fd})
    except CommandNotFound:
        writemsg(_("portage: stty: command not found\n"), noiselevel=-1)


class EOutput:
    """
    Performs fancy terminal formatting for status and informational messages.

    The provided methods produce identical terminal output to the eponymous
    functions in the shell script C{/sbin/functions.sh} and also accept
    identical parameters.

    This is not currently a drop-in replacement however, as the output-related
    functions in C{/sbin/functions.sh} are oriented for use mainly by system
    init scripts and ebuilds and their output can be customized via certain
    C{RC_*} environment variables (see C{/etc/conf.d/rc}). B{EOutput} is not
    customizable in this manner since it's intended for more general uses.
    Likewise, no logging is provided.

    @ivar quiet: Specifies if output should be silenced.
    @type quiet: BooleanType
    @ivar term_columns: Width of terminal in characters. Defaults to the value
            specified by the shell's C{COLUMNS} variable, else to the queried tty
            size, else to C{80}.
    @type term_columns: IntType
    """

    def __init__(self, quiet=False):
        self.__last_e_cmd = ""
        self.__last_e_len = 0
        self.quiet = quiet
        lines, columns = get_term_size()
        if columns <= 0:
            columns = 80
        self.term_columns = columns
        sys.stdout.flush()
        sys.stderr.flush()

    def _write(self, f, s):
        # avoid potential UnicodeEncodeError
        writemsg(s, noiselevel=-1, fd=f)

    def __eend(self, caller, errno, msg):
        if errno == 0:
            status_brackets = (
                colorize("BRACKET", "[ ")
                + colorize("GOOD", "ok")
                + colorize("BRACKET", " ]")
            )
        else:
            status_brackets = (
                colorize("BRACKET", "[ ")
                + colorize("BAD", "!!")
                + colorize("BRACKET", " ]")
            )
            if msg:
                if caller == "eend":
                    self.eerror(msg[0])
                elif caller == "ewend":
                    self.ewarn(msg[0])
        if self.__last_e_cmd != "ebegin":
            self.__last_e_len = 0
        if not self.quiet:
            out = sys.stdout
            self._write(
                out,
                "%*s%s\n"
                % ((self.term_columns - self.__last_e_len - 7), "", status_brackets),
            )

    def ebegin(self, msg):
        """
        Shows a message indicating the start of a process.

        @param msg: A very brief (shorter than one line) description of the
                starting process.
        @type msg: StringType
        """
        msg += " ..."
        if not self.quiet:
            self.einfon(msg)
        self.__last_e_len = len(msg) + 3
        self.__last_e_cmd = "ebegin"

    def eend(self, errno, *msg):
        """
        Indicates the completion of a process, optionally displaying a message
        via L{eerror} if the process's exit status isn't C{0}.

        @param errno: A standard UNIX C{errno} code returned by processes upon
                exit.
        @type errno: IntType
        @param msg: I{(optional)} An error message, typically a standard UNIX
                error string corresponding to C{errno}.
        @type msg: StringType
        """
        if not self.quiet:
            self.__eend("eend", errno, msg)
        self.__last_e_cmd = "eend"

    def eerror(self, msg):
        """
        Shows an error message.

        @param msg: A very brief (shorter than one line) error message.
        @type msg: StringType
        """
        out = sys.stderr
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("ERR", " * ") + msg + "\n")
        self.__last_e_cmd = "eerror"

    def einfo(self, msg):
        """
        Shows an informative message terminated with a newline.

        @param msg: A very brief (shorter than one line) informative message.
        @type msg: StringType
        """
        out = sys.stdout
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("INFO", " * ") + msg + "\n")
        self.__last_e_cmd = "einfo"

    def einfon(self, msg):
        """
        Shows an informative message terminated without a newline.

        @param msg: A very brief (shorter than one line) informative message.
        @type msg: StringType
        """
        out = sys.stdout
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("INFO", " * ") + msg)
        self.__last_e_cmd = "einfon"

    def eqawarn(self, msg):
        """
        Shows a QA warning.

        @param msg: A very brief (shorter than one line) warning message.
        @type msg: StringType
        """
        out = sys.stderr
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("QAWARN", " * ") + msg + "\n")
        self.__last_e_cmd = "ewarn"

    def elog(self, msg):
        """
        Shows a logged informative message terminated with a newline.

        @param msg: A very brief (shorter than one line) informative message.
        @type msg: StringType
        """
        out = sys.stdout
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("LOG", " * ") + msg + "\n")
        self.__last_e_cmd = "elog"

    def ewarn(self, msg):
        """
        Shows a warning message.

        @param msg: A very brief (shorter than one line) warning message.
        @type msg: StringType
        """
        out = sys.stderr
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("WARN", " * ") + msg + "\n")
        self.__last_e_cmd = "ewarn"

    def ewend(self, errno, *msg):
        """
        Indicates the completion of a process, optionally displaying a message
        via L{ewarn} if the process's exit status isn't C{0}.

        @param errno: A standard UNIX C{errno} code returned by processes upon
                exit.
        @type errno: IntType
        @param msg: I{(optional)} A warning message, typically a standard UNIX
                error string corresponding to C{errno}.
        @type msg: StringType
        """
        if not self.quiet:
            self.__eend("ewend", errno, msg)
        self.__last_e_cmd = "ewend"


class ProgressBar:
    """The interface is copied from the ProgressBar class from the EasyDialogs
    module (which is Mac only)."""

    def __init__(self, title=None, maxval=0, label=None, max_desc_length=25):
        self._title = title or ""
        self._maxval = maxval
        self._label = label or ""
        self._curval = 0
        self._desc = ""
        self._desc_max_length = max_desc_length
        self._set_desc()

    @property
    def curval(self):
        """
        The current value (of type integer or long integer) of the progress
        bar. The normal access methods coerce curval between 0 and maxval. This
        attribute should not be altered directly.
        """
        return self._curval

    @property
    def maxval(self):
        """
        The maximum value (of type integer or long integer) of the progress
        bar; the progress bar (thermometer style) is full when curval equals
        maxval. If maxval is 0, the bar will be indeterminate (barber-pole).
        This attribute should not be altered directly.
        """
        return self._maxval

    def title(self, newstr):
        """Sets the text in the title bar of the progress dialog to newstr."""
        self._title = newstr
        self._set_desc()

    def label(self, newstr):
        """Sets the text in the progress box of the progress dialog to newstr."""
        self._label = newstr
        self._set_desc()

    def _set_desc(self):
        self._desc = "%s%s" % (
            "%s: " % self._title if self._title else "",
            "%s" % self._label if self._label else "",
        )
        if len(self._desc) > self._desc_max_length:  # truncate if too long
            self._desc = "%s..." % self._desc[: self._desc_max_length - 3]
        if len(self._desc):
            self._desc = self._desc.ljust(self._desc_max_length)

    def set(self, value, maxval=None):
        """
        Sets the progress bar's curval to value, and also maxval to max if the
        latter is provided. value is first coerced between 0 and maxval. The
        thermometer bar is updated to reflect the changes, including a change
        from indeterminate to determinate or vice versa.
        """
        if maxval is not None:
            self._maxval = maxval
        if value < 0:
            value = 0
        elif value > self._maxval:
            value = self._maxval
        self._curval = value

    def inc(self, n=1):
        """Increments the progress bar's curval by n, or by 1 if n is not
        provided. (Note that n may be negative, in which case the effect is a
        decrement.) The progress bar is updated to reflect the change. If the
        bar is indeterminate, this causes one ``spin'' of the barber pole. The
        resulting curval is coerced between 0 and maxval if incrementing causes
        it to fall outside this range.
        """
        self.set(self._curval + n)


class TermProgressBar(ProgressBar):
    """A tty progress bar similar to wget's."""

    def __init__(self, fd=sys.stdout, **kwargs):
        ProgressBar.__init__(self, **kwargs)
        lines, self.term_columns = get_term_size(fd)
        self.file = fd
        self._min_columns = 11
        self._max_columns = 80
        # for indeterminate mode, ranges from 0.0 to 1.0
        self._position = 0.0

    def set(self, value, maxval=None):
        ProgressBar.set(self, value, maxval=maxval)
        self._display_image(self._create_image())

    def _display_image(self, image):
        self.file.write("\r")
        self.file.write(image)
        self.file.flush()

    def _create_image(self):
        cols = self.term_columns
        if cols > self._max_columns:
            cols = self._max_columns
        min_columns = self._min_columns
        curval = self._curval
        maxval = self._maxval
        position = self._position
        percentage_str_width = 5
        square_brackets_width = 2
        if cols < percentage_str_width:
            return ""
        bar_space = cols - percentage_str_width - square_brackets_width - 1
        if self._desc:
            bar_space -= self._desc_max_length

        if maxval == 0:
            max_bar_width = bar_space - 3
            _percent = "".ljust(percentage_str_width)
            if cols < min_columns:
                return ""
            if position <= 0.5:
                offset = 2 * position
            else:
                offset = 2 * (1 - position)
            delta = 0.5 / max_bar_width
            position += delta
            if position >= 1.0:
                position = 0.0
            # make sure it touches the ends
            if 1.0 - position < delta:
                position = 1.0
            if position < 0.5 and 0.5 - position < delta:
                position = 0.5
            self._position = position
            bar_width = int(offset * max_bar_width)
            image = "%s%s%s" % (
                self._desc,
                _percent,
                "["
                + (bar_width * " ")
                + "<=>"
                + ((max_bar_width - bar_width) * " ")
                + "]",
            )
            return image

        percentage = 100 * curval // maxval
        max_bar_width = bar_space - 1
        _percent = ("%d%% " % percentage).rjust(percentage_str_width)
        image = "%s%s" % (self._desc, _percent)

        if cols < min_columns:
            return image
        offset = curval / maxval
        bar_width = int(offset * max_bar_width)
        image = (
            image
            + "["
            + (bar_width * "=")
            + ">"
            + ((max_bar_width - bar_width) * " ")
            + "]"
        )
        return image


_color_map_loaded = False


def _init(config_root="/"):
    """
    Load color.map from the given config_root. This is called automatically
    on first access of the codes or _styles attributes (unless it has already
    been called for some other reason).
    """

    global _color_map_loaded, codes, _styles
    if _color_map_loaded:
        return

    _color_map_loaded = True
    codes = object.__getattribute__(codes, "_attr")
    _styles = object.__getattribute__(_styles, "_attr")

    for k, v in codes.items():
        codes[k] = _unicode_decode(v)

    for k, v in _styles.items():
        _styles[k] = _unicode_decode(v)

    try:
        _parse_color_map(
            config_root=config_root,
            onerror=lambda e: writemsg("%s\n" % str(e), noiselevel=-1),
        )
    except FileNotFound:
        pass
    except PermissionDenied as e:
        writemsg(_("Permission denied: '%s'\n") % str(e), noiselevel=-1)
        del e
    except PortageException as e:
        writemsg("%s\n" % str(e), noiselevel=-1)
        del e


class _LazyInitColorMap(portage.proxy.objectproxy.ObjectProxy):

    __slots__ = ("_attr",)

    def __init__(self, attr):
        portage.proxy.objectproxy.ObjectProxy.__init__(self)
        object.__setattr__(self, "_attr", attr)

    def _get_target(self):
        _init()
        return object.__getattribute__(self, "_attr")


codes = _LazyInitColorMap(codes)
_styles = _LazyInitColorMap(_styles)



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractPollTask import AbstractPollTask

from portage import os
from portage.cache.mappings import slot_dict_class


class FifoIpcDaemon(AbstractPollTask):

    __slots__ = ("input_fifo", "output_fifo", "_files")

    _file_names = ("pipe_in",)
    _files_dict = slot_dict_class(_file_names, prefix="")

    def _start(self):
        self._files = self._files_dict()

        # File streams are in unbuffered mode since we do atomic
        # read and write of whole pickles.
        self._files.pipe_in = os.open(self.input_fifo, os.O_RDONLY | os.O_NONBLOCK)

        self.scheduler.add_reader(self._files.pipe_in, self._input_handler)

        self._registered = True

    def _reopen_input(self):
        """
        Re-open the input stream, in order to suppress
        POLLHUP events (bug #339976).
        """
        self.scheduler.remove_reader(self._files.pipe_in)
        os.close(self._files.pipe_in)
        self._files.pipe_in = os.open(self.input_fifo, os.O_RDONLY | os.O_NONBLOCK)

        self.scheduler.add_reader(self._files.pipe_in, self._input_handler)

    def _cancel(self):
        if self.returncode is None:
            self.returncode = 1
        self._unregister()
        # notify exit listeners
        self._async_wait()

    def _input_handler(self):
        raise NotImplementedError(self)

    def _unregister(self):
        """
        Unregister from the scheduler and close open files.
        """

        self._registered = False

        if self._files is not None:
            for f in self._files.values():
                self.scheduler.remove_reader(f)
                os.close(f)
            self._files = None



# Copyright 2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["bin_entry_point"]

import re
import sys

from portage.const import PORTAGE_BIN_PATH
from portage import os


def bin_entry_point():
    """
    Adjust sys.argv[0] to point to a script in PORTAGE_BIN_PATH, and
    then execute the script, in order to implement entry_points when
    portage has been installed by pip.
    """
    script_path = os.path.join(PORTAGE_BIN_PATH, os.path.basename(sys.argv[0]))
    if os.access(script_path, os.X_OK):
        with open(script_path, "rt") as f:
            shebang = f.readline()
        python_match = re.search(r"/python[\d\.]*\s+([^/]*)\s+$", shebang)
        if python_match:
            sys.argv = [
                os.path.join(os.path.dirname(sys.argv[0]), "python"),
                python_match.group(1),
                script_path,
            ] + sys.argv[1:]
            os.execvp(sys.argv[0], sys.argv)
        sys.argv[0] = script_path
        os.execvp(sys.argv[0], sys.argv)
    else:
        print("File not found:", script_path, file=sys.stderr)
        return 127



# Copyright 2012-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from _emerge.AsynchronousTask import AsynchronousTask
from _emerge.PollScheduler import PollScheduler


class AsyncScheduler(AsynchronousTask, PollScheduler):
    def __init__(self, max_jobs=None, max_load=None, **kwargs):
        AsynchronousTask.__init__(self)
        PollScheduler.__init__(self, **kwargs)

        if max_jobs is None:
            max_jobs = 1
        self._max_jobs = max_jobs
        self._max_load = None if max_load is True else max_load
        self._error_count = 0
        self._running_tasks = set()
        self._remaining_tasks = True
        self._loadavg_check_id = None

    @property
    def scheduler(self):
        """
        Provides compatibility with the AsynchronousTask.scheduler attribute.
        """
        return self._event_loop

    def _poll(self):
        if not (self._is_work_scheduled() or self._keep_scheduling()):
            if self._error_count > 0:
                self.returncode = 1
            else:
                self.returncode = os.EX_OK
            self._async_wait()
        return self.returncode

    def _cancel(self):
        self._terminated.set()
        self._termination_check()

    def _terminate_tasks(self):
        for task in list(self._running_tasks):
            task.cancel()

    def _next_task(self):
        raise NotImplementedError(self)

    def _keep_scheduling(self):
        return self._remaining_tasks and not self._terminated.is_set()

    def _running_job_count(self):
        return len(self._running_tasks)

    def _schedule_tasks(self):
        while self._keep_scheduling() and self._can_add_job():
            try:
                task = self._next_task()
            except StopIteration:
                self._remaining_tasks = False
            else:
                self._running_tasks.add(task)
                task.scheduler = self._sched_iface
                task.addExitListener(self._task_exit)
                task.start()

        if self._loadavg_check_id is not None:
            self._loadavg_check_id.cancel()
            self._loadavg_check_id = self._event_loop.call_later(
                self._loadavg_latency, self._schedule
            )

        # Triggers cleanup and exit listeners if there's nothing left to do.
        self.poll()

    def _task_exit(self, task):
        self._running_tasks.discard(task)
        if task.returncode != os.EX_OK:
            self._error_count += 1
        self._schedule()

    def _start(self):
        if (
            self._max_load is not None
            and self._loadavg_latency is not None
            and (self._max_jobs is True or self._max_jobs > 1)
        ):
            # We have to schedule periodically, in case the load
            # average has changed since the last call.
            self._loadavg_check_id = self._event_loop.call_later(
                self._loadavg_latency, self._schedule
            )
        self._schedule()

    def _cleanup(self):
        super(AsyncScheduler, self)._cleanup()
        if self._loadavg_check_id is not None:
            self._loadavg_check_id.cancel()
            self._loadavg_check_id = None

    def _async_wait(self):
        """
        Override _async_wait to call self._cleanup().
        """
        self._cleanup()
        super(AsyncScheduler, self)._async_wait()



# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import pty
import shutil
import socket
import tempfile

from portage import os
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from _emerge.PipeReader import PipeReader


class PipeReaderTestCase(TestCase):

    _use_array = False
    _echo_cmd = "echo -n '%s'"

    def test_pipe(self):
        def make_pipes():
            return os.pipe(), None

        self._do_test(make_pipes)

    def test_pty_device(self):
        def make_pipes():
            try:
                return pty.openpty(), None
            except EnvironmentError:
                self.skipTest("pty not available")

        self._do_test(make_pipes)

    def test_domain_socket(self):
        def make_pipes():
            read_end, write_end = socket.socketpair()
            return (read_end.detach(), write_end.detach()), None

        self._do_test(make_pipes)

    def test_named_pipe(self):
        def make_pipes():
            tempdir = tempfile.mkdtemp()
            fifo_path = os.path.join(tempdir, "fifo")
            os.mkfifo(fifo_path)
            return (
                (
                    os.open(fifo_path, os.O_NONBLOCK | os.O_RDONLY),
                    os.open(fifo_path, os.O_NONBLOCK | os.O_WRONLY),
                ),
                functools.partial(shutil.rmtree, tempdir),
            )

        self._do_test(make_pipes)

    def _testPipeReader(self, master_fd, slave_fd, test_string):
        """
        Use a poll loop to read data from a pipe and assert that
        the data written to the pipe is identical to the data
        read from the pipe.
        """

        # WARNING: It is very important to use unbuffered mode here,
        # in order to avoid issue 5380 with python3.
        master_file = os.fdopen(master_fd, "rb", 0)
        scheduler = global_event_loop()

        consumer = PipeReader(
            input_files={"producer": master_file},
            _use_array=self._use_array,
            scheduler=scheduler,
        )
        consumer.start()

        producer = scheduler.run_until_complete(
            asyncio.create_subprocess_exec(
                "bash",
                "-c",
                self._echo_cmd % test_string,
                stdout=slave_fd,
                loop=scheduler,
            )
        )

        os.close(slave_fd)
        scheduler.run_until_complete(producer.wait())
        scheduler.run_until_complete(consumer.async_wait())

        self.assertEqual(producer.returncode, os.EX_OK)
        self.assertEqual(consumer.returncode, os.EX_OK)

        return consumer.getvalue().decode("ascii", "replace")

    def _do_test(self, make_pipes):
        for x in (1, 2, 5, 6, 7, 8, 2**5, 2**10, 2**12, 2**13, 2**14):
            test_string = x * "a"
            (read_end, write_end), cleanup = make_pipes()
            try:
                output = self._testPipeReader(read_end, write_end, test_string)
                self.assertEqual(
                    test_string, output, "x = %s, len(output) = %s" % (x, len(output))
                )
            finally:
                if cleanup is not None:
                    cleanup()


class PipeReaderArrayTestCase(PipeReaderTestCase):

    _use_array = True
    # sleep allows reliable triggering of the failure mode on fast computers
    _echo_cmd = "sleep 0.1 ; echo -n '%s'"

    def __init__(self, *args, **kwargs):
        super(PipeReaderArrayTestCase, self).__init__(*args, **kwargs)
        # https://bugs.python.org/issue5380
        # https://bugs.pypy.org/issue956
        self.todo = True



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class DisjunctiveDependOrderTestCase(TestCase):
    def testDisjunctiveDependOrderTestCase(self):
        ebuilds = {
            "virtual/jre-1.8": {
                "EAPI": "6",
                "SLOT": "1.8",
                "RDEPEND": "|| ( dev-java/oracle-jre-bin:1.8 virtual/jdk:1.8 )",
            },
            "virtual/jdk-1.8": {
                "EAPI": "6",
                "SLOT": "1.8",
                "RDEPEND": "|| ( dev-java/icedtea:8 dev-java/oracle-jdk-bin:1.8 )",
            },
            "dev-java/icedtea-3.6": {
                "SLOT": "8",
            },
            "dev-java/oracle-jdk-bin-1.8": {
                "SLOT": "1.8",
            },
            "dev-java/oracle-jre-bin-1.8": {
                "SLOT": "1.8",
            },
            "dev-db/hsqldb-1.8": {
                "DEPEND": "virtual/jdk",
                "RDEPEND": "virtual/jre",
            },
        }

        binpkgs = {
            "dev-db/hsqldb-1.8": {
                "DEPEND": "virtual/jdk",
                "RDEPEND": "virtual/jre",
            },
        }

        test_cases = (
            # Test bug 639346, where a redundant jre implementation
            # was pulled in because DEPEND was evaluated after
            # RDEPEND.
            ResolverPlaygroundTestCase(
                ["dev-db/hsqldb"],
                success=True,
                mergelist=[
                    "dev-java/icedtea-3.6",
                    "virtual/jdk-1.8",
                    "virtual/jre-1.8",
                    "dev-db/hsqldb-1.8",
                ],
            ),
            # The jdk is not needed with --usepkg, so the jre should
            # be preferred in this case.
            ResolverPlaygroundTestCase(
                ["dev-db/hsqldb"],
                options={"--usepkg": True},
                success=True,
                mergelist=[
                    "dev-java/oracle-jre-bin-1.8",
                    "virtual/jre-1.8",
                    "[binary]dev-db/hsqldb-1.8",
                ],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    debug=False,
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()



# Copyright 2009-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import errno
import itertools
import stat
import sys

from portage.const import PORTAGE_BIN_PATH, PORTAGE_PYM_PATH, PORTAGE_PYM_PACKAGES
from portage.tests import TestCase
from portage.tests.lint.metadata import module_metadata, script_metadata
from portage import os
from portage import _encodings
from portage import _unicode_decode, _unicode_encode


class CompileModulesTestCase(TestCase):
    def testCompileModules(self):
        iters = [
            os.walk(os.path.join(PORTAGE_PYM_PATH, x)) for x in PORTAGE_PYM_PACKAGES
        ]
        iters.append(os.walk(PORTAGE_BIN_PATH))

        for parent, _dirs, files in itertools.chain(*iters):
            parent = _unicode_decode(parent, encoding=_encodings["fs"], errors="strict")
            for x in files:
                x = _unicode_decode(x, encoding=_encodings["fs"], errors="strict")
                if x[-4:] in (".pyc", ".pyo"):
                    continue
                x = os.path.join(parent, x)
                st = os.lstat(x)
                if not stat.S_ISREG(st.st_mode):
                    continue

                bin_path = os.path.relpath(x, PORTAGE_BIN_PATH)
                mod_path = os.path.relpath(x, PORTAGE_PYM_PATH)

                meta = module_metadata.get(mod_path) or script_metadata.get(bin_path)
                if meta:
                    req_py = tuple(
                        int(x) for x in meta.get("required_python", "0.0").split(".")
                    )
                    if sys.version_info < req_py:
                        continue

                do_compile = False
                if x[-3:] == ".py":
                    do_compile = True
                else:
                    # Check for python shebang.
                    try:
                        with open(
                            _unicode_encode(
                                x, encoding=_encodings["fs"], errors="strict"
                            ),
                            "rb",
                        ) as f:
                            line = _unicode_decode(
                                f.readline(),
                                encoding=_encodings["content"],
                                errors="replace",
                            )
                    except IOError as e:
                        # Some tests create files that are unreadable by the
                        # user (by design), so ignore EACCES issues.
                        if e.errno != errno.EACCES:
                            raise
                        continue
                    if line[:2] == "#!" and "python" in line:
                        do_compile = True
                if do_compile:
                    with open(
                        _unicode_encode(x, encoding=_encodings["fs"], errors="strict"),
                        "rb",
                    ) as f:
                        compile(f.read(), x, "exec")



# portage: news management code
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = [
    "NewsManager",
    "NewsItem",
    "DisplayRestriction",
    "DisplayProfileRestriction",
    "DisplayKeywordRestriction",
    "DisplayInstalledRestriction",
    "count_unread_news",
    "display_news_notifications",
]

from collections import OrderedDict

import fnmatch
import io
import logging
import os as _os
import re
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.const import NEWS_LIB_PATH
from portage.util import (
    apply_secpass_permissions,
    ensure_dirs,
    grabfile,
    normalize_path,
    write_atomic,
    writemsg_level,
)
from portage.data import portage_gid
from portage.dep import isvalidatom
from portage.localization import _
from portage.locks import lockfile, unlockfile
from portage.output import colorize
from portage.exception import (
    InvalidLocation,
    OperationNotPermitted,
    PermissionDenied,
    ReadOnlyFileSystem,
)


class NewsManager:
    """
    This object manages GLEP 42 style news items.  It will cache news items
    that have previously shown up and notify users when there are relevant news
    items that apply to their packages that the user has not previously read.

    Creating a news manager requires:
    root - typically ${ROOT} see man make.conf and man emerge for details
    news_path - path to news items; usually $REPODIR/metadata/news
    unread_path - path to the news.repoid.unread file; this helps us track news items

    """

    def __init__(self, portdb, vardb, news_path, unread_path, language_id="en"):
        self.news_path = news_path
        self.unread_path = unread_path
        self.language_id = language_id
        self.config = vardb.settings
        self.vdb = vardb
        self.portdb = portdb

        # GLEP 42 says:
        #   All news item related files should be root owned and in the
        #   portage group with the group write (and, for directories,
        #   execute) bits set. News files should be world readable.
        self._uid = int(self.config["PORTAGE_INST_UID"])
        self._gid = portage_gid
        self._file_mode = 0o0064
        self._dir_mode = 0o0074
        self._mode_mask = 0o0000

        portdir = portdb.repositories.mainRepoLocation()
        profiles_base = None
        if portdir is not None:
            profiles_base = os.path.join(portdir, "profiles", os.path.sep)
        profile_path = None
        if profiles_base is not None and portdb.settings.profile_path:
            profile_path = normalize_path(
                os.path.realpath(portdb.settings.profile_path)
            )
            if profile_path.startswith(profiles_base):
                profile_path = profile_path[len(profiles_base) :]
        self._profile_path = profile_path

    def _unread_filename(self, repoid):
        return os.path.join(self.unread_path, f"news-{repoid}.unread")

    def _skip_filename(self, repoid):
        return os.path.join(self.unread_path, f"news-{repoid}.skip")

    def _news_dir(self, repoid):
        repo_path = self.portdb.getRepositoryPath(repoid)
        if repo_path is None:
            raise AssertionError(_(f"Invalid repoID: {repoid}"))
        return os.path.join(repo_path, self.news_path)

    def updateItems(self, repoid):
        """
        Figure out which news items from NEWS_PATH are both unread and relevant to
        the user (according to the GLEP 42 standards of relevancy).  Then add these
        items into the news.repoid.unread file.
        """

        # Ensure that the unread path exists and is writable.

        try:
            ensure_dirs(
                self.unread_path,
                uid=self._uid,
                gid=self._gid,
                mode=self._dir_mode,
                mask=self._mode_mask,
            )
        except (OperationNotPermitted, PermissionDenied):
            return

        if not os.access(self.unread_path, os.W_OK):
            return

        news_dir = self._news_dir(repoid)
        try:
            news = _os.listdir(
                _unicode_encode(news_dir, encoding=_encodings["fs"], errors="strict")
            )
        except OSError:
            return

        skip_filename = self._skip_filename(repoid)
        unread_filename = self._unread_filename(repoid)
        unread_lock = lockfile(unread_filename, wantnewlockfile=1)
        try:
            try:
                unread = set(grabfile(unread_filename))
                unread_orig = unread.copy()
                skip = set(grabfile(skip_filename))
                skip_orig = skip.copy()
            except PermissionDenied:
                return

            for itemid in news:
                try:
                    itemid = _unicode_decode(
                        itemid, encoding=_encodings["fs"], errors="strict"
                    )
                except UnicodeDecodeError:
                    itemid = _unicode_decode(
                        itemid, encoding=_encodings["fs"], errors="replace"
                    )
                    writemsg_level(
                        _("!!! Invalid encoding in news item name: '%s'\n") % itemid,
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    continue

                if itemid in skip:
                    continue
                filename = os.path.join(
                    news_dir, itemid, f"{itemid}.{self.language_id}.txt"
                )
                if not os.path.isfile(filename):
                    continue
                item = NewsItem(filename, itemid)
                if not item.isValid():
                    continue
                if item.isRelevant(
                    profile=self._profile_path, config=self.config, vardb=self.vdb
                ):
                    unread.add(item.name)
                    skip.add(item.name)

            if unread != unread_orig:
                write_atomic(unread_filename, "".join(f"{x}\n" for x in sorted(unread)))
                apply_secpass_permissions(
                    unread_filename,
                    uid=self._uid,
                    gid=self._gid,
                    mode=self._file_mode,
                    mask=self._mode_mask,
                )

            if skip != skip_orig:
                write_atomic(skip_filename, "".join(f"{x}\n" for x in sorted(skip)))
                apply_secpass_permissions(
                    skip_filename,
                    uid=self._uid,
                    gid=self._gid,
                    mode=self._file_mode,
                    mask=self._mode_mask,
                )

        finally:
            unlockfile(unread_lock)

    def getUnreadItems(self, repoid, update=False):
        """
        Determine if there are unread relevant items in news.repoid.unread.
        If there are unread items return their number.
        If update is specified, updateNewsItems( repoid ) will be called to
        check for new items.
        """

        if update:
            self.updateItems(repoid)

        unread_filename = self._unread_filename(repoid)
        unread_lock = None
        try:
            unread_lock = lockfile(unread_filename, wantnewlockfile=1)
        except (
            InvalidLocation,
            OperationNotPermitted,
            PermissionDenied,
            ReadOnlyFileSystem,
        ):
            pass
        try:
            try:
                return len(grabfile(unread_filename))
            except PermissionDenied:
                return 0
        finally:
            if unread_lock:
                unlockfile(unread_lock)


_formatRE = re.compile(r"News-Item-Format:\s*([^\s]*)\s*$")
_installedRE = re.compile("Display-If-Installed:(.*)\n")
_profileRE = re.compile("Display-If-Profile:(.*)\n")
_keywordRE = re.compile("Display-If-Keyword:(.*)\n")
_valid_profile_RE = re.compile(r"^[^*]+(/\*)?$")


class NewsItem:
    """
    This class encapsulates a GLEP 42 style news item.
    It's purpose is to wrap parsing of these news items such that portage can determine
    whether a particular item is 'relevant' or not.  This requires parsing the item
    and determining 'relevancy restrictions'; these include "Display if Installed" or
    "display if arch: x86" and so forth.

    Creation of a news item involves passing in the path to the particular news item.
    """

    def __init__(self, path, name):
        """
        For a given news item we only want if it path is a file.
        """
        self.path = path
        self.name = name
        self._parsed = False
        self._valid = True

    def isRelevant(self, vardb, config, profile):
        """
        This function takes a dict of keyword arguments; one should pass in any
        objects need to do to lookups (like what keywords we are on, what profile,
        and a vardb so we can look at installed packages).
        Each restriction will pluck out the items that are required for it to match
        or raise a ValueError exception if the required object is not present.

        Restrictions of the form Display-X are OR'd with like-restrictions;
        otherwise restrictions are AND'd.  any_match is the ORing and
        all_match is the ANDing.
        """

        if not self._parsed:
            self.parse()

        if not len(self.restrictions):
            return True

        kwargs = {"vardb": vardb, "config": config, "profile": profile}

        all_match = all(
            restriction.checkRestriction(**kwargs)
            for values in self.restrictions.values()
            for restriction in values
        )

        return all_match

    def isValid(self):
        if not self._parsed:
            self.parse()
        return self._valid

    def parse(self):
        with io.open(
            _unicode_encode(self.path, encoding=_encodings["fs"], errors="strict"),
            mode="r",
            encoding=_encodings["content"],
            errors="replace",
        ) as f:
            lines = f.readlines()
        self.restrictions = {}
        invalids = []
        news_format = None

        # Look for News-Item-Format
        for i, line in enumerate(lines):
            format_match = _formatRE.match(line)
            if format_match is not None:
                news_format = format_match.group(1)
                if fnmatch.fnmatch(news_format, "[12].*"):
                    break
                invalids.append((i + 1, line.rstrip("\n")))

        if news_format is None:
            invalids.append((0, "News-Item-Format unspecified"))
        else:
            # Parse the rest
            for i, line in enumerate(lines):
                # Optimization to ignore regex matches on lines that
                # will never match
                if not line.startswith("D"):
                    continue
                restricts = {
                    _installedRE: DisplayInstalledRestriction,
                    _profileRE: DisplayProfileRestriction,
                    _keywordRE: DisplayKeywordRestriction,
                }
                for regex, restriction in restricts.items():
                    match = regex.match(line)
                    if match:
                        restrict = restriction(match.groups()[0].strip(), news_format)
                        if not restrict.isValid():
                            invalids.append((i + 1, line.rstrip("\n")))
                        else:
                            self.restrictions.setdefault(id(restriction), []).append(
                                restrict
                            )
                        continue

        if invalids:
            self._valid = False
            msg = [
                _(f"Invalid news item: {self.path}"),
                *(_(f"  line {lineno}: {line}") for lineno, line in invalids),
            ]
            writemsg_level(
                "".join(f"!!! {x}\n" for x in msg), level=logging.ERROR, noiselevel=-1
            )

        self._parsed = True


class DisplayRestriction:
    """
    A base restriction object representing a restriction of display.
    news items may have 'relevancy restrictions' preventing them from
    being important.  In this case we need a manner of figuring out if
    a particular item is relevant or not.  If any of it's restrictions
    are met, then it is displayed
    """

    def isValid(self):
        return True

    def checkRestriction(self, **kwargs):
        raise NotImplementedError("Derived class should override this method")


class DisplayProfileRestriction(DisplayRestriction):
    """
    A profile restriction where a particular item shall only be displayed
    if the user is running a specific profile.
    """

    def __init__(self, profile, news_format):
        self.profile = profile
        self.format = news_format

    def isValid(self):
        return (
            not fnmatch.fnmatch(self.format, "1.*")
            or "*" not in self.profile
            and not fnmatch.fnmatch(self.format, "2.*")
            or _valid_profile_RE.match(self.profile)
        )

    def checkRestriction(self, **kwargs):
        if fnmatch.fnmatch(self.format, "2.*") and self.profile.endswith("/*"):
            return kwargs["profile"].startswith(self.profile[:-1])
        return kwargs["profile"] == self.profile


class DisplayKeywordRestriction(DisplayRestriction):
    """
    A keyword restriction where a particular item shall only be displayed
    if the user is running a specific keyword.
    """

    def __init__(self, keyword, news_format):
        self.keyword = keyword
        self.format = news_format

    def checkRestriction(self, **kwargs):
        return kwargs["config"].get("ARCH", "") == self.keyword


class DisplayInstalledRestriction(DisplayRestriction):
    """
    An Installation restriction where a particular item shall only be displayed
    if the user has that item installed.
    """

    def __init__(self, atom, news_format):
        self.atom = atom
        self.format = news_format

    def isValid(self):
        if fnmatch.fnmatch(self.format, "1.*"):
            return isvalidatom(self.atom, eapi="0")
        if fnmatch.fnmatch(self.format, "2.*"):
            return isvalidatom(self.atom, eapi="5")
        return isvalidatom(self.atom)

    def checkRestriction(self, **kwargs):
        return kwargs["vardb"].match(self.atom)


def count_unread_news(portdb, vardb, repos=None, update=True):
    """
    Returns a dictionary mapping repos to integer counts of unread news items.
    By default, this will scan all repos and check for new items that have
    appeared since the last scan.

    @param portdb: an ebuild database
    @type portdb: pordbapi
    @param vardb: an installed package database
    @type vardb: vardbapi
    @param repos: names of repos to scan (None means to scan all available repos)
    @type repos: list or None
    @param update: check for new items (default is True)
    @type update: boolean
    @rtype: dict
    @return: dictionary mapping repos to integer counts of unread news items
    """

    NEWS_PATH = os.path.join("metadata", "news")
    UNREAD_PATH = os.path.join(vardb.settings["EROOT"], NEWS_LIB_PATH, "news")
    news_counts = OrderedDict()
    if repos is None:
        repos = portdb.getRepositories()

    permission_msgs = set()
    for repo in repos:
        try:
            manager = NewsManager(portdb, vardb, NEWS_PATH, UNREAD_PATH)
            count = manager.getUnreadItems(repo, update=True)
        except PermissionDenied as e:
            # NOTE: The NewsManager typically handles permission errors by
            # returning silently, so PermissionDenied won't necessarily be
            # raised even if we do trigger a permission error above.
            msg = f"Permission denied: '{e}'\n"
            if msg in permission_msgs:
                pass
            else:
                permission_msgs.add(msg)
                writemsg_level(msg, level=logging.ERROR, noiselevel=-1)
            news_counts[repo] = 0
        else:
            news_counts[repo] = count

    return news_counts


def display_news_notifications(news_counts):
    """
    Display a notification for unread news items, using a dictionary mapping
    repos to integer counts, like that returned from count_unread_news().
    """
    newsReaderDisplay = False
    for repo, count in news_counts.items():
        if count > 0:
            if not newsReaderDisplay:
                newsReaderDisplay = True
                print()
            print(colorize("WARN", " * IMPORTANT:"), end=" ")
            print(f"{count} news items need reading for repository '{repo}'.")

    if newsReaderDisplay:
        print(colorize("WARN", " *"), end=" ")
        print(f"Use {colorize('GOOD', 'eselect news read')} to view new items.\n")



# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("KeywordsManager",)

import warnings

import portage
from portage import os
from portage.dep import ExtendedAtomDict
from portage.localization import _
from portage.package.ebuild._config.helper import ordered_by_atom_specificity
from portage.repository.config import allow_profile_repo_deps
from portage.util import grabdict_package, stack_lists
from portage.versions import _pkg_str


class KeywordsManager:
    """Manager class to handle keywords processing and validation"""

    def __init__(
        self, profiles, abs_user_config, user_config=True, global_accept_keywords=""
    ):
        self._pkeywords_list = []
        rawpkeywords = [
            grabdict_package(
                os.path.join(x.location, "package.keywords"),
                recursive=x.portage1_directories,
                verify_eapi=True,
                eapi=x.eapi,
                eapi_default=None,
                allow_repo=allow_profile_repo_deps(x),
                allow_build_id=x.allow_build_id,
            )
            for x in profiles
        ]
        for pkeyworddict in rawpkeywords:
            if not pkeyworddict:
                # Omit non-existent files from the stack.
                continue
            cpdict = {}
            for k, v in pkeyworddict.items():
                cpdict.setdefault(k.cp, {})[k] = v
            self._pkeywords_list.append(cpdict)
        self._pkeywords_list = tuple(self._pkeywords_list)

        self._p_accept_keywords = []
        raw_p_accept_keywords = [
            grabdict_package(
                os.path.join(x.location, "package.accept_keywords"),
                recursive=x.portage1_directories,
                verify_eapi=True,
                eapi=x.eapi,
                eapi_default=None,
                allow_repo=allow_profile_repo_deps(x),
            )
            for x in profiles
        ]
        for d in raw_p_accept_keywords:
            if not d:
                # Omit non-existent files from the stack.
                continue
            cpdict = {}
            for k, v in d.items():
                cpdict.setdefault(k.cp, {})[k] = tuple(v)
            self._p_accept_keywords.append(cpdict)
        self._p_accept_keywords = tuple(self._p_accept_keywords)

        self.pkeywordsdict = ExtendedAtomDict(dict)

        if user_config:
            user_accept_kwrds_path = os.path.join(
                abs_user_config, "package.accept_keywords"
            )
            user_kwrds_path = os.path.join(abs_user_config, "package.keywords")
            pkgdict = grabdict_package(
                user_kwrds_path,
                recursive=1,
                allow_wildcard=True,
                allow_repo=True,
                verify_eapi=False,
                allow_build_id=True,
            )

            if pkgdict and portage._internal_caller:
                warnings.warn(
                    _("%s is deprecated, use %s instead")
                    % (user_kwrds_path, user_accept_kwrds_path),
                    UserWarning,
                )

            for k, v in grabdict_package(
                user_accept_kwrds_path,
                recursive=1,
                allow_wildcard=True,
                allow_repo=True,
                verify_eapi=False,
                allow_build_id=True,
            ).items():
                pkgdict.setdefault(k, []).extend(v)

            accept_keywords_defaults = global_accept_keywords.split()
            accept_keywords_defaults = tuple(
                "~" + keyword
                for keyword in accept_keywords_defaults
                if keyword[:1] not in "~-"
            )
            for k, v in pkgdict.items():
                # default to ~arch if no specific keyword is given
                if not v:
                    v = accept_keywords_defaults
                else:
                    v = tuple(v)
                self.pkeywordsdict.setdefault(k.cp, {})[k] = v

    def getKeywords(self, cpv, slot, keywords, repo):
        try:
            cpv.slot
        except AttributeError:
            pkg = _pkg_str(cpv, slot=slot, repo=repo)
        else:
            pkg = cpv
        cp = pkg.cp
        keywords = [[x for x in keywords.split() if x != "-*"]]
        for pkeywords_dict in self._pkeywords_list:
            cpdict = pkeywords_dict.get(cp)
            if cpdict:
                pkg_keywords = ordered_by_atom_specificity(cpdict, pkg)
                if pkg_keywords:
                    keywords.extend(pkg_keywords)
        return stack_lists(keywords, incremental=True)

    def isStable(self, pkg, global_accept_keywords, backuped_accept_keywords):
        mygroups = self.getKeywords(pkg, None, pkg._metadata["KEYWORDS"], None)
        pgroups = global_accept_keywords.split()

        unmaskgroups = self.getPKeywords(pkg, None, None, global_accept_keywords)
        pgroups.extend(unmaskgroups)

        egroups = backuped_accept_keywords.split()

        if unmaskgroups or egroups:
            pgroups = self._getEgroups(egroups, pgroups)
        else:
            pgroups = set(pgroups)

        if self._getMissingKeywords(pkg, pgroups, mygroups):
            return False

        # If replacing all keywords with unstable variants would mask the
        # package, then it's considered stable for the purposes of
        # use.stable.mask/force interpretation. For unstable configurations,
        # this guarantees that the effective use.force/mask settings for a
        # particular ebuild do not change when that ebuild is stabilized.
        unstable = []
        for kw in mygroups:
            if kw[:1] != "~":
                kw = "~" + kw
            unstable.append(kw)

        return bool(self._getMissingKeywords(pkg, pgroups, set(unstable)))

    def getMissingKeywords(
        self,
        cpv,
        slot,
        keywords,
        repo,
        global_accept_keywords,
        backuped_accept_keywords,
    ):
        """
        Take a package and return a list of any KEYWORDS that the user may
        need to accept for the given package. If the KEYWORDS are empty
        and the ** keyword has not been accepted, the returned list will
        contain ** alone (in order to distinguish from the case of "none
        missing").

        @param cpv: The package name (for package.keywords support)
        @type cpv: String
        @param slot: The 'SLOT' key from the raw package metadata
        @type slot: String
        @param keywords: The 'KEYWORDS' key from the raw package metadata
        @type keywords: String
        @param global_accept_keywords: The current value of ACCEPT_KEYWORDS
        @type global_accept_keywords: String
        @param backuped_accept_keywords: ACCEPT_KEYWORDS from the backup env
        @type backuped_accept_keywords: String
        @rtype: List
        @return: A list of KEYWORDS that have not been accepted.
        """

        mygroups = self.getKeywords(cpv, slot, keywords, repo)
        # Repoman may modify this attribute as necessary.
        pgroups = global_accept_keywords.split()

        unmaskgroups = self.getPKeywords(cpv, slot, repo, global_accept_keywords)
        pgroups.extend(unmaskgroups)

        # Hack: Need to check the env directly here as otherwise stacking
        # doesn't work properly as negative values are lost in the config
        # object (bug #139600)
        egroups = backuped_accept_keywords.split()

        if unmaskgroups or egroups:
            pgroups = self._getEgroups(egroups, pgroups)
        else:
            pgroups = set(pgroups)

        return self._getMissingKeywords(cpv, pgroups, mygroups)

    def getRawMissingKeywords(self, cpv, slot, keywords, repo, global_accept_keywords):
        """
        Take a package and return a list of any KEYWORDS that the user may
        need to accept for the given package. If the KEYWORDS are empty,
        the returned list will contain ** alone (in order to distinguish
        from the case of "none missing").  This DOES NOT apply any user config
        package.accept_keywords acceptance.

        @param cpv: The package name (for package.keywords support)
        @type cpv: String
        @param slot: The 'SLOT' key from the raw package metadata
        @type slot: String
        @param keywords: The 'KEYWORDS' key from the raw package metadata
        @type keywords: String
        @param global_accept_keywords: The current value of ACCEPT_KEYWORDS
        @type global_accept_keywords: String
        @rtype: List
        @return: lists of KEYWORDS that have not been accepted
        and the keywords it looked for.
        """

        mygroups = self.getKeywords(cpv, slot, keywords, repo)
        pgroups = global_accept_keywords.split()
        pgroups = set(pgroups)
        return self._getMissingKeywords(cpv, pgroups, mygroups)

    @staticmethod
    def _getEgroups(egroups, mygroups):
        """gets any keywords defined in the environment

        @param backuped_accept_keywords: ACCEPT_KEYWORDS from the backup env
        @type backuped_accept_keywords: String
        @rtype: List
        @return: list of KEYWORDS that have been accepted
        """
        mygroups = list(mygroups)
        mygroups.extend(egroups)
        inc_pgroups = set()
        for x in mygroups:
            if x[:1] == "-":
                if x == "-*":
                    inc_pgroups.clear()
                else:
                    inc_pgroups.discard(x[1:])
            else:
                inc_pgroups.add(x)
        return inc_pgroups

    @staticmethod
    def _getMissingKeywords(cpv, pgroups, mygroups):
        """Determines the missing keywords

        @param pgroups: The pkg keywords accepted
        @type pgroups: list
        @param mygroups: The ebuild keywords
        @type mygroups: list
        """
        match = False
        hasstable = False
        hastesting = False
        for gp in mygroups:
            if gp == "*":
                match = True
                break
            elif gp == "~*":
                hastesting = True
                for x in pgroups:
                    if x[:1] == "~":
                        match = True
                        break
                if match:
                    break
            elif gp in pgroups:
                match = True
                break
            elif gp.startswith("~"):
                hastesting = True
            elif not gp.startswith("-"):
                hasstable = True
        if not match and (
            (hastesting and "~*" in pgroups)
            or (hasstable and "*" in pgroups)
            or "**" in pgroups
        ):
            match = True
        if match:
            missing = []
        else:
            if not mygroups:
                # If KEYWORDS is empty then we still have to return something
                # in order to distinguish from the case of "none missing".
                mygroups = ["**"]
            missing = mygroups
        return missing

    def getPKeywords(self, cpv, slot, repo, global_accept_keywords):
        """Gets any package.keywords settings for cp for the given
        cpv, slot and repo

        @param cpv: The package name (for package.keywords support)
        @type cpv: String
        @param slot: The 'SLOT' key from the raw package metadata
        @type slot: String
        @param keywords: The 'KEYWORDS' key from the raw package metadata
        @type keywords: String
        @param global_accept_keywords: The current value of ACCEPT_KEYWORDS
        @type global_accept_keywords: String
        @param backuped_accept_keywords: ACCEPT_KEYWORDS from the backup env
        @type backuped_accept_keywords: String
        @rtype: List
        @return: list of KEYWORDS that have been accepted
        """

        pgroups = global_accept_keywords.split()
        try:
            cpv.slot
        except AttributeError:
            cpv = _pkg_str(cpv, slot=slot, repo=repo)
        cp = cpv.cp

        unmaskgroups = []
        if self._p_accept_keywords:
            accept_keywords_defaults = tuple(
                "~" + keyword for keyword in pgroups if keyword[:1] not in "~-"
            )
            for d in self._p_accept_keywords:
                cpdict = d.get(cp)
                if cpdict:
                    pkg_accept_keywords = ordered_by_atom_specificity(cpdict, cpv)
                    if pkg_accept_keywords:
                        for x in pkg_accept_keywords:
                            if not x:
                                x = accept_keywords_defaults
                            unmaskgroups.extend(x)

        pkgdict = self.pkeywordsdict.get(cp)
        if pkgdict:
            pkg_accept_keywords = ordered_by_atom_specificity(pkgdict, cpv)
            if pkg_accept_keywords:
                for x in pkg_accept_keywords:
                    unmaskgroups.extend(x)
        return unmaskgroups



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import warnings

from _emerge.Package import Package
from _emerge.PackageVirtualDbapi import PackageVirtualDbapi
from _emerge.resolver.DbapiProvidesIndex import PackageDbapiProvidesIndex

import portage
from portage import os
from portage.const import VDB_PATH
from portage.dbapi.vartree import vartree
from portage.dep._slot_operator import find_built_slot_operator_atoms
from portage.eapi import _get_eapi_attrs
from portage.exception import InvalidData, InvalidDependString
from portage.update import grab_updates, parse_updates, update_dbentries
from portage.versions import _pkg_str


class FakeVardbGetPath:
    """
    Implements the vardbapi.getpath() method which is used in error handling
    code for the Package class and vartree.get_provide().
    """

    def __init__(self, vardb):
        self.settings = vardb.settings

    def __call__(self, cpv, filename=None):
        path = os.path.join(self.settings["EROOT"], VDB_PATH, cpv)
        if filename is not None:
            path = os.path.join(path, filename)
        return path


class _DynamicDepsNotApplicable(Exception):
    pass


class FakeVartree(vartree):
    """This is implements an in-memory copy of a vartree instance that provides
    all the interfaces required for use by the depgraph.  The vardb is locked
    during the constructor call just long enough to read a copy of the
    installed package information.  This allows the depgraph to do it's
    dependency calculations without holding a lock on the vardb.  It also
    allows things like vardb global updates to be done in memory so that the
    user doesn't necessarily need write access to the vardb in cases where
    global updates are necessary (updates are performed when necessary if there
    is not a matching ebuild in the tree). Instances of this class are not
    populated until the sync() method is called."""

    def __init__(
        self,
        root_config,
        pkg_cache=None,
        pkg_root_config=None,
        dynamic_deps=False,
        ignore_built_slot_operator_deps=False,
        soname_deps=False,
    ):
        self._root_config = root_config
        self._dynamic_deps = dynamic_deps
        self._ignore_built_slot_operator_deps = ignore_built_slot_operator_deps
        if pkg_root_config is None:
            pkg_root_config = self._root_config
        self._pkg_root_config = pkg_root_config
        if pkg_cache is None:
            pkg_cache = {}
        real_vartree = root_config.trees["vartree"]
        self._real_vardb = real_vartree.dbapi
        portdb = root_config.trees["porttree"].dbapi
        self.settings = real_vartree.settings
        mykeys = list(real_vartree.dbapi._aux_cache_keys)
        if "_mtime_" not in mykeys:
            mykeys.append("_mtime_")
        self._db_keys = mykeys
        self._pkg_cache = pkg_cache
        self.dbapi = PackageVirtualDbapi(real_vartree.settings)
        if soname_deps:
            self.dbapi = PackageDbapiProvidesIndex(self.dbapi)
        self.dbapi.getpath = FakeVardbGetPath(self.dbapi)
        self.dbapi._aux_cache_keys = set(self._db_keys)

        # Initialize variables needed for lazy cache pulls of the live ebuild
        # metadata.  This ensures that the vardb lock is released ASAP, without
        # being delayed in case cache generation is triggered.
        self._aux_get = self.dbapi.aux_get
        self._match = self.dbapi.match
        if dynamic_deps:
            self.dbapi.aux_get = self._aux_get_wrapper
            self.dbapi.match = self._match_wrapper
        self._aux_get_history = set()
        self._portdb_keys = Package._dep_keys + ("EAPI", "KEYWORDS")
        self._portdb = portdb
        self._global_updates = None

    @property
    def root(self):
        warnings.warn(
            "The root attribute of "
            "_emerge.FakeVartree.FakeVartree"
            " is deprecated. Use "
            "settings['ROOT'] instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.settings["ROOT"]

    def _match_wrapper(self, cpv, use_cache=1):
        """
        Make sure the metadata in Package instances gets updated for any
        cpv that is returned from a match() call, since the metadata can
        be accessed directly from the Package instance instead of via
        aux_get().
        """
        matches = self._match(cpv, use_cache=use_cache)
        for cpv in matches:
            if cpv in self._aux_get_history:
                continue
            self._aux_get_wrapper(cpv, [])
        return matches

    def _aux_get_wrapper(self, cpv, wants, myrepo=None):
        if cpv in self._aux_get_history:
            return self._aux_get(cpv, wants)
        self._aux_get_history.add(cpv)

        # This raises a KeyError to the caller if appropriate.
        pkg = self.dbapi._cpv_map[cpv]

        try:
            live_metadata = dict(
                zip(
                    self._portdb_keys,
                    self._portdb.aux_get(cpv, self._portdb_keys, myrepo=pkg.repo),
                )
            )
        except (KeyError, portage.exception.PortageException):
            live_metadata = None

        self._apply_dynamic_deps(pkg, live_metadata)

        return self._aux_get(cpv, wants)

    def _apply_dynamic_deps(self, pkg, live_metadata):

        try:
            if live_metadata is None:
                raise _DynamicDepsNotApplicable()
            # Use the metadata from the installed instance if the EAPI
            # of either instance is unsupported, since if the installed
            # instance has an unsupported or corrupt EAPI then we don't
            # want to attempt to do complex operations such as execute
            # pkg_config, pkg_prerm or pkg_postrm phases. If both EAPIs
            # are supported then go ahead and use the live_metadata, in
            # order to respect dep updates without revision bump or EAPI
            # bump, as in bug #368725.
            if not (
                portage.eapi_is_supported(live_metadata["EAPI"])
                and portage.eapi_is_supported(pkg.eapi)
            ):
                raise _DynamicDepsNotApplicable()

            # preserve built slot/sub-slot := operator deps
            built_slot_operator_atoms = None
            if (
                not self._ignore_built_slot_operator_deps
                and _get_eapi_attrs(pkg.eapi).slot_operator
            ):
                try:
                    built_slot_operator_atoms = find_built_slot_operator_atoms(pkg)
                except InvalidDependString:
                    pass

            if built_slot_operator_atoms:
                live_eapi_attrs = _get_eapi_attrs(live_metadata["EAPI"])
                if not live_eapi_attrs.slot_operator:
                    raise _DynamicDepsNotApplicable()
                for k, v in built_slot_operator_atoms.items():
                    live_metadata[k] += " " + " ".join(str(atom) for atom in v)

            self.dbapi.aux_update(pkg.cpv, live_metadata)
        except _DynamicDepsNotApplicable:
            if self._global_updates is None:
                self._global_updates = grab_global_updates(self._portdb)

            # Bypass _aux_get_wrapper, since calling that
            # here would trigger infinite recursion.
            aux_keys = Package._dep_keys + self.dbapi._pkg_str_aux_keys
            aux_dict = dict(zip(aux_keys, self._aux_get(pkg.cpv, aux_keys)))
            perform_global_updates(pkg.cpv, aux_dict, self.dbapi, self._global_updates)

    def dynamic_deps_preload(self, pkg, metadata):
        if metadata is not None:
            metadata = dict((k, metadata.get(k, "")) for k in self._portdb_keys)
        self._apply_dynamic_deps(pkg, metadata)
        self._aux_get_history.add(pkg.cpv)

    def cpv_discard(self, pkg):
        """
        Discard a package from the fake vardb if it exists.
        """
        old_pkg = self.dbapi.get(pkg)
        if old_pkg is not None:
            self.dbapi.cpv_remove(old_pkg)
            self._pkg_cache.pop(old_pkg, None)
            self._aux_get_history.discard(old_pkg.cpv)

    def sync(self, acquire_lock=1):
        """
        Call this method to synchronize state with the real vardb
        after one or more packages may have been installed or
        uninstalled.
        """
        locked = False
        try:
            if acquire_lock and os.access(self._real_vardb._dbroot, os.W_OK):
                self._real_vardb.lock()
                locked = True
            self._sync()
        finally:
            if locked:
                self._real_vardb.unlock()

        # Populate the old-style virtuals using the cached values.
        # Skip the aux_get wrapper here, to avoid unwanted
        # cache generation.
        try:
            self.dbapi.aux_get = self._aux_get
            self.settings._populate_treeVirtuals_if_needed(self)
        finally:
            if self._dynamic_deps:
                self.dbapi.aux_get = self._aux_get_wrapper

    def _sync(self):

        real_vardb = self._root_config.trees["vartree"].dbapi
        current_cpv_set = frozenset(real_vardb.cpv_all())
        pkg_vardb = self.dbapi

        # Remove any packages that have been uninstalled.
        for pkg in list(pkg_vardb):
            if pkg.cpv not in current_cpv_set:
                self.cpv_discard(pkg)

        # Validate counters and timestamps.
        slot_counters = {}
        root_config = self._pkg_root_config
        validation_keys = ["COUNTER", "_mtime_"]
        for cpv in current_cpv_set:

            pkg_hash_key = Package._gen_hash_key(
                cpv=cpv, installed=True, root_config=root_config, type_name="installed"
            )
            pkg = pkg_vardb.get(pkg_hash_key)
            if pkg is not None:
                counter, mtime = real_vardb.aux_get(cpv, validation_keys)
                try:
                    counter = int(counter)
                except ValueError:
                    counter = 0

                if counter != pkg.counter or mtime != pkg.mtime:
                    self.cpv_discard(pkg)
                    pkg = None

            if pkg is None:
                pkg = self._pkg(cpv)

            other_counter = slot_counters.get(pkg.slot_atom)
            if other_counter is not None:
                if other_counter > pkg.counter:
                    continue

            slot_counters[pkg.slot_atom] = pkg.counter
            pkg_vardb.cpv_inject(pkg)

        real_vardb.flush_cache()

    def _pkg(self, cpv):
        """
        The RootConfig instance that will become the Package.root_config
        attribute can be overridden by the FakeVartree pkg_root_config
        constructory argument, since we want to be consistent with the
        depgraph._pkg() method which uses a specially optimized
        RootConfig that has a FakeVartree instead of a real vartree.
        """
        pkg = Package(
            cpv=cpv,
            built=True,
            installed=True,
            metadata=zip(self._db_keys, self._real_vardb.aux_get(cpv, self._db_keys)),
            root_config=self._pkg_root_config,
            type_name="installed",
        )

        self._pkg_cache[pkg] = pkg
        return pkg


def grab_global_updates(portdb):
    retupdates = {}

    for repo_name in portdb.getRepositories():
        repo = portdb.getRepositoryPath(repo_name)
        updpath = os.path.join(repo, "profiles", "updates")
        if not os.path.isdir(updpath):
            continue

        try:
            rawupdates = grab_updates(updpath)
        except portage.exception.DirectoryNotFound:
            rawupdates = []
        upd_commands = []
        for mykey, mystat, mycontent in rawupdates:
            commands, errors = parse_updates(mycontent)
            upd_commands.extend(commands)
        retupdates[repo_name] = upd_commands

    master_repo = portdb.repositories.mainRepo()
    if master_repo is not None:
        master_repo = master_repo.name
    if master_repo in retupdates:
        retupdates["DEFAULT"] = retupdates[master_repo]

    return retupdates


def perform_global_updates(mycpv, aux_dict, mydb, myupdates):
    try:
        pkg = _pkg_str(mycpv, metadata=aux_dict, settings=mydb.settings)
    except InvalidData:
        return
    aux_dict = dict((k, aux_dict[k]) for k in Package._dep_keys)
    try:
        mycommands = myupdates[pkg.repo]
    except KeyError:
        try:
            mycommands = myupdates["DEFAULT"]
        except KeyError:
            return

    if not mycommands:
        return

    updates = update_dbentries(mycommands, aux_dict, parent=pkg)
    if updates:
        mydb.aux_update(mycpv, updates)
